from obt._dep_build import BaseBuilder
from obt._dep_impl import require
from obt import pathtools, path, _globals
from obt.command import Command
from collections.abc import Callable

###############################################################################

class CustomStep:
  def __init__(self,name,funktor):
    super().__init__()
    self._name = name
    self._funktor = funktor 

###############################################################################

class CustomBuilder(BaseBuilder):
  ###########################################
  def __init__(self,name):
    super().__init__(name)
    ##################################
    self._envvars = dict()
    self._parallelism=1.0
    self._cleanbuildcommands = list()
    self._incrbuildcommands = list()
    self._installcommands = list()
    self._builddir = path.builds()/name
    self._invokeBeforeBuild = None

    if _globals.tryBoolOption("serial"):
      self._parallelism=0.0
  ###########################################
  def requires(self,deplist):
    self._deps += deplist
  ###########################################
  def setEnvVar(self,key,value):
    self._envvars[key] = value
  ###########################################
  def setEnvVars(self,othdict):
    for k in othdict:
      self._envvars[k] = othdict[k]
  ###########################################
  def _run_commands(self,cmdlist):
    if len(cmdlist)>0:
      for cmd in cmdlist:
        if isinstance(cmd,Command):
          print("cmd<%s>"%cmd)
          retc = cmd.exec()
          if retc!=0:
            assert(False)
        elif isinstance(cmd,CustomStep):
          retc = cmd._funktor()
          if retc==False:
            assert(False)
        elif isinstance(cmd,Callable):
          if cmd()==False:
            assert(False)
    return True
  ###########################################
  def build(self,srcdir,blddir,wrkdir,incremental=False):
    
    if(self._invokeBeforeBuild!=None):
      self._invokeBeforeBuild()
      
    print("srcdir<%s>"%srcdir)
    print("blddir<%s>"%blddir)
    print("incremental<%s>"%incremental)
    print("deps<%s>"%self._deps)
    ok2build = require(self._deps)
    print("ok2build<%d>"%int(ok2build))
    if not ok2build:
      return False
    ###################################
    if incremental:
    ###################################
      print("doing incremental build...")
      print(self._incrbuildcommands)
      pathtools.mkdir(self._builddir,clean=False,parents=True)
      pathtools.chdir(self._builddir)
      return self._run_commands(self._incrbuildcommands)
    ###################################
    else: # clean build
    ###################################
      print("doing clean build...")
      print(self._cleanbuildcommands)
      pathtools.mkdir(self._builddir,clean=False,parents=True)
      pathtools.chdir(self._builddir)
      return self._run_commands(self._cleanbuildcommands)
  ###########################################
  def install(self,blddir):
    pathtools.chdir(self._builddir)
    return self._run_commands(self._installcommands)
  ###########################################



import sys, threading, time, json, random, os, base64
from pathlib import Path
from yarl import URL
from orkbb import BranchStatus,WorkerStatus, DEBUGFORMAT
import _watcher, _zmqssh,zmq
################################################################################
# MasterConfigWorker
################################################################################
class Worker(object):
  ############################################
  def __init__(self,name,configobj):
    self.name = name
    self.config = configobj
    self.thread = None
    self.variantlist = list()
    self.status = WorkerStatus.INIT
    self.scheduler = None
    self.curvaridx = 0
    self.curvarname = None
    self.curvariant = None
    self.buildstatus = None
    self.connected = False
    self.jobindex = 0
  ############################################
  def start(self,sched):
    self.scheduler = sched
  ############################################
  def curVariant(self):
    varnames = sorted([item.name for item in self.variantlist])
    numvarnames = len(varnames)
    varname = varnames[self.curvaridx]
    vari = self.scheduler.config.variants[varname]
    return vari
  ############################################
  def advanceVariant(self):
    numvarnames = len(self.variantlist)
    self.curvaridx += 1
    if self.curvaridx > (numvarnames-1):
      self.curvaridx = 0
    self.jobindex += 1
  ############################################
  def curRepo(self):
    return self.curVariant().repo
  ############################################
  def curBranch(self):
      vari = self.curVariant()
      repo = self.curRepo()
      repo.watcher.update_branch_info()
      branch_obj = repo.watcher.get_branch(vari.branchname)
      return branch_obj
  ############################################
  def onWaiting(self):
    reply = {"state": "wait"}
    vari = self.curVariant()
    repo = self.curRepo()
    branch_obj = self.curBranch()
    if branch_obj != None:
      reply["jobindex"] = self.jobindex
    return reply
  ############################################
  def onInitializing(self):
    reply = {"state": "wait"}
    vari = self.curVariant()
    repo = self.curRepo()
    branch_obj = self.curBranch()
    if branch_obj != None:
      self.curvarname = vari.name
      self.curvariant = self.scheduler.config.variants[vari.name]
      self.curvariant.prvbuildstatus=self.curvariant.curbuildstatus
      self.curvariant.curbuildstatus = BranchStatus.BUILDING
      self.curvariant.head_sha = branch_obj.last_sha
      self.curvariant.shaurl = branch_obj.last_commit_url
      self.status = WorkerStatus.FETCHING
      reply = {"state": "init"}
      reply["sha"] = branch_obj.last_sha
      reply["variant"] = str(self.curvarname)
      reply["platform"] = vari.platform
      reply["branchname"] = str(vari.branchname)
      self.scheduler.onWorkerStatus(self)
    return reply
  ############################################
  def onFetching(self):
    reply = {"state": "wait"}
    vari = self.curVariant()
    repo = self.curRepo()
    branch_obj = self.curBranch()
    if branch_obj != None:
        reply = {"state": "fetch"}
        reply["variant"] = str(self.curvarname)
        reply["sha"] = branch_obj.last_sha
        reply["repourl"] = str(repo.giturl)
        reply["reponame"] = repo.name
        reply["skipLFS"] = repo.skipLFS
        self.status = WorkerStatus.BUILDING
        self.scheduler.onWorkerStatus(self)
    return reply
  ############################################
  def onBuilding(self):
    reply = {"state": "wait"}
    if self.status == WorkerStatus.BUILDING:
      vari = self.curvariant
      repo = vari.repo
      reply = {"state": "build"}
      reply["workingdir"] = repo.workingdir
      reply["cmd"] = "build"
      reply["commandlist"] = repo.commandlist
      reply["subdir"] = repo.subdir
    return reply
  ############################################
  def onReporting(self,message):
    reply = {"state": "wait"}
    if self.status == WorkerStatus.BUILDING:
      self.curvariant.buildtime = message["buildtime"]
      if message["status"]=="OK":
        self.curvariant.curbuildstatus = BranchStatus.PASSED
      else:
        self.curvariant.curbuildstatus = BranchStatus.FAILED
      #######################
      ba = bytearray(base64.b64decode(message["log_html_enc"]))
      f = open(self.curvariant.log_html, 'wb')
      f.write(ba)
      f.close()
      #######################
      # advance variant on worker
      #######################
      self.curvarname = None
      self.curvariant = None
      self.status = WorkerStatus.WAIT
      self.scheduler.onWorkerStatus(self)
      self.advanceVariant()
      reply = {"state": "report"}
    return reply
  ############################################
  def __repr__(self):
    return str(self.__dict__)
################################################################################
# MasterConfigRepo
################################################################################
class Repo(object):
  ############################################
  def __init__(self,config,repo_key):
    configobj = config.config_dict["repos"][repo_key]
    self.name = repo_key
    self.config = config
    self.giturl = configobj["giturl"]
    self.subdir = configobj["subdir"]
    self.skipLFS = configobj["skipLFS"]
    self.workingdir = configobj["workingdir"]
    self.commandlist = configobj["commandlist"]
    self.enabled = True
    if "enable" in configobj:
      self.enabled = configobj["enable"]
    self.watcher = None
  def poll(self):
    pass
  ############################################
  def startWatcher(self):
    if self.enabled:
      print("INIT REPOOBJ repo_key<%s>"%self.name)
      print("INIT REPOOBJ giturl<%s>"%self.giturl)
      print("INIT REPOOBJ subdir<%s>"%self.subdir)
      self.watcher = _watcher.RepoWatcher(self.config,self)
  ############################################
  def __repr__(self):
    return str(self.__dict__)
################################################################################
# MasterConfigVariant
################################################################################
class Variant(object):
  ############################################
  def __init__(self,configobj,name,outputdir,repo,branchname):
    self.name = name
    self.enabled = True
    self.schedule = configobj["schedule"]
    self.workers = configobj["workers"]
    self.branchname = branchname
    self.platform = configobj["platform"]
    if "enable" in configobj:
      self.enabled = configobj["enable"]
    self.repo = repo
    self.repourl = repo.giturl
    self.outputdir = outputdir/name
    self.outrelpath = Path(name)
    self.index_html= self.outrelpath/"index.html"
    self.log_html = self.outrelpath/"log_stdout.html"
    self.status_svg= self.outrelpath/"status.svg"
    self.curbuildstatus = BranchStatus.INIT
    self.prvbuildstatus = BranchStatus.INIT
    self.branchurl = repo.giturl
    self.buildtime="---"
    self.head_sha="---"
    self.shaurl = "---"
  ############################################
  def updateRepoData(self):
    if self.repo.watcher!=None:
      branch = self.repo.watcher.getBranch(self.branchname)
      if branch!=None:
        self.branchurl = branch.remote_giturl
  ############################################
  def __repr__(self):
    return str(self.__dict__)
################################################################################
# MasterConfig
################################################################################
class Config(object):
  ###################################################################
  def __init__(self,configfile,outputdir):
    PROJTEXT = ""
    self.repos = {}
    self.variants = {}
    self.workers = {}
    self.master_name = "???"
    self.user_branch_list=[]
    self.skipLFS = False
    self.sleeptime = 3600
    self.workingdir= "$REPODIR"
    self.outputdir= outputdir
    self.repobasedir = self.outputdir/"repo"
    self.buildsbasedir = self.outputdir/"builds"
    print("repobasedir<%s>"%self.repobasedir)
    print("self.buildsbasedir<%s>"%self.buildsbasedir)
    print("remove old build")
    os.system("rm -rf %s"%self.repobasedir)
    os.system("mkdir -p %s"%self.repobasedir)
    os.system("mkdir -p %s"%self.buildsbasedir)
    #######################################
    with open(configfile,"r") as f:
      PROJTEXT = f.read()
    self.config_dict = json.loads(PROJTEXT)
    #######################################
    self.master_name = self.config_dict["master_name"]
    self.master_bind_addr = self.config_dict["master_bind_addr"]
    self.master_bind_port = self.config_dict["master_bind_port"]
    self.http_bind_addr = self.config_dict["http_bind_addr"]
    self.http_bind_port = self.config_dict["http_bind_port"]
    #######################################
    for worker_key in self.config_dict["workers"].keys():
      worker = self.config_dict["workers"][worker_key]
      self.workers[worker_key] = Worker(worker_key,worker)
    #######################################
    for repo_key in self.config_dict["repos"].keys():
      self.repos[repo_key] = Repo(self,repo_key)
    #######################################
    for vari_key in self.config_dict["variants"].keys():
      vari = self.config_dict["variants"][vari_key]
      repo = self.repos[vari["repo"]]
      for branch in vari["branches"]:
        qualname = Path(vari_key)/branch
        outdir = outputdir/qualname
        variant = Variant(vari,qualname,outdir,repo,branch)
        if variant.enabled:
          self.variants[qualname] = variant
          for w in variant.workers:
            worker = self.workers[w]
            worker.variantlist += [variant]
  ###################################################################
  def start(self):
    for repo_key in self.config_dict["repos"].keys():
      self.repos[repo_key].startWatcher()



#!/usr/bin/env python3
"""
C++ Database File Management Tool
Search, list, and retrieve source files from the C++ database
"""

import argparse
import sys
from pathlib import Path
from typing import List, Dict, Any

from obt.cpp_database_v2 import CppDatabaseV2
import obt.path as obt_path


def list_files(db: CppDatabaseV2, limit: int = None) -> None:
    """List all files in the database"""
    files = db.list_source_files(limit)
    
    if not files:
        print("No files found in database")
        return
    
    print(f"Found {len(files)} files:")
    print("=" * 80)
    
    for file_info in files:
        size_kb = file_info['file_size'] / 1024 if file_info['file_size'] else 0
        print(f"{file_info['relative_path']:<50} {size_kb:>8.1f} KB  {file_info['updated_at']}")


def search_files(db: CppDatabaseV2, pattern: str) -> None:
    """Search for files by pattern"""
    files = db.search_source_files(pattern)
    
    if not files:
        print(f"No files found matching pattern: {pattern}")
        return
    
    print(f"Found {len(files)} files matching '{pattern}':")
    print("=" * 80)
    
    for file_info in files:
        size_kb = file_info['file_size'] / 1024 if file_info['file_size'] else 0
        print(f"{file_info['relative_path']:<50} {size_kb:>8.1f} KB  {file_info['updated_at']}")


def get_file_by_name(db: CppDatabaseV2, filename: str) -> None:
    """Get files by exact filename"""
    files = db.get_source_file_by_name(filename)
    
    if not files:
        print(f"No files found with name: {filename}")
        return
    
    if len(files) == 1:
        file_info = files[0]
        print(f"File: {file_info['file_path']}")
        print(f"Size: {file_info['file_size']} bytes")
        print(f"Updated: {file_info['updated_at']}")
        print("=" * 80)
    else:
        print(f"Found {len(files)} files with name '{filename}':")
        print("=" * 80)
        for i, file_info in enumerate(files, 1):
            size_kb = file_info['file_size'] / 1024 if file_info['file_size'] else 0
            print(f"{i}. {file_info['file_path']}")
            print(f"   Size: {size_kb:.1f} KB, Updated: {file_info['updated_at']}")


def show_file_content(db: CppDatabaseV2, file_path: str, show_raw: bool = True, 
                     show_preprocessed: bool = False, line_numbers: bool = False) -> None:
    """Show file content"""
    file_info = db.get_source_file(file_path)
    
    if not file_info:
        print(f"File not found in database: {file_path}")
        return
    
    if show_raw:
        print(f"=== RAW SOURCE: {file_info['relative_path']} ===")
        content = file_info['raw_source']
        if line_numbers:
            for i, line in enumerate(content.split('\n'), 1):
                print(f"{i:4d}: {line}")
        else:
            print(content)
    
    if show_preprocessed and file_info['preprocessed_source']:
        print(f"\n=== PREPROCESSED SOURCE: {file_info['relative_path']} ===")
        content = file_info['preprocessed_source']
        if line_numbers:
            for i, line in enumerate(content.split('\n'), 1):
                print(f"{i:4d}: {line}")
        else:
            print(content)
    elif show_preprocessed:
        print(f"No preprocessed source available for: {file_info['relative_path']}")


def main():
    parser = argparse.ArgumentParser(description='C++ Database File Management Tool')
    parser.add_argument('-d', '--database', help='Database file path (default: auto-detect)')
    
    subparsers = parser.add_subparsers(dest='command', help='Available commands')
    
    # List command
    list_parser = subparsers.add_parser('list', help='List all files in database')
    list_parser.add_argument('--limit', type=int, help='Limit number of files shown')
    
    # Search command
    search_parser = subparsers.add_parser('search', help='Search files by pattern')
    search_parser.add_argument('pattern', help='Search pattern (wildcards supported)')
    
    # Find command
    find_parser = subparsers.add_parser('find', help='Find files by exact filename')
    find_parser.add_argument('filename', help='Exact filename to find')
    
    # Show command
    show_parser = subparsers.add_parser('show', help='Show file content')
    show_parser.add_argument('file_path', help='Full file path to show')
    show_parser.add_argument('--raw', action='store_true', default=True, 
                           help='Show raw source (default)')
    show_parser.add_argument('--preprocessed', action='store_true', 
                           help='Show preprocessed source')
    show_parser.add_argument('--both', action='store_true', 
                           help='Show both raw and preprocessed')
    show_parser.add_argument('-n', '--line-numbers', action='store_true', 
                           help='Show line numbers')
    
    args = parser.parse_args()
    
    if not args.command:
        parser.print_help()
        return
    
    # Determine database path
    if args.database:
        db_path = Path(args.database)
    else:
        # Auto-detect database
        db_path = obt_path.stage() / "cpp_db_v2_orkid.db"
        if not db_path.exists():
            print(f"Database not found at {db_path}")
            print("Run database build first or specify database path with -d")
            return
    
    # Create database connection
    db = CppDatabaseV2(str(db_path))
    
    # Execute command
    try:
        if args.command == 'list':
            list_files(db, args.limit)
        
        elif args.command == 'search':
            search_files(db, args.pattern)
        
        elif args.command == 'find':
            get_file_by_name(db, args.filename)
        
        elif args.command == 'show':
            show_raw = args.raw or not (args.preprocessed or args.both)
            show_preprocessed = args.preprocessed or args.both
            show_file_content(db, args.file_path, show_raw, show_preprocessed, args.line_numbers)
        
    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)


if __name__ == '__main__':
    main()


#!/usr/bin/env python3
"""
Test that validates C++ class parsing by comparing database entities 
against Clang AST parsing of preprocessed source code. 
This provides ground truth validation using Clang's authoritative C++ parser.
"""

import sys
import os
from pathlib import Path
from collections import defaultdict
import json
import clang.cindex
from clang.cindex import CursorKind, AccessSpecifier

# Add obt to path
import obt.path
import obt.deco
from obt.cpp_database_v2 import CppDatabaseV2
from obt.cpp_entities_v2 import EntityType, MemberType

deco = obt.deco.Deco()

class ClangClassValidator:
    def __init__(self, db_path):
        self.db = CppDatabaseV2(db_path)
        self.clang_index = clang.cindex.Index.create()
        self.errors = []
        self.warnings = []
        self.stats = {
            'classes_checked': 0,
            'methods_found': 0,
            'methods_missing': 0,
            'methods_extra': 0,
            'fields_found': 0,
            'fields_missing': 0,
            'fields_extra': 0,
            'overloads_detected': 0,
            'access_mismatches': 0,
        }
        
    def get_access_string(self, access_spec):
        """Convert Clang AccessSpecifier to string."""
        if access_spec == AccessSpecifier.PUBLIC:
            return 'public'
        elif access_spec == AccessSpecifier.PROTECTED:
            return 'protected'
        elif access_spec == AccessSpecifier.PRIVATE:
            return 'private'
        return 'private'  # Default for classes
    
    def get_method_signature(self, cursor):
        """Extract method signature from cursor."""
        # Get parameter types
        params = []
        for child in cursor.get_children():
            if child.kind == CursorKind.PARM_DECL:
                param_type = child.type.spelling
                param_name = child.spelling or ""
                params.append(param_type)
        
        param_str = ", ".join(params)
        signature = f"{cursor.spelling}({param_str})"
        
        # Add const qualifier if present
        if cursor.is_const_method():
            signature += " const"
            
        return signature
    
    def find_class_cursor(self, root_cursor, class_name):
        """Find class cursor by name in AST."""
        for cursor in root_cursor.walk_preorder():
            if cursor.kind in [CursorKind.CLASS_DECL, CursorKind.STRUCT_DECL]:
                if cursor.spelling == class_name:
                    # Make sure it's a definition, not just declaration
                    if cursor.is_definition():
                        return cursor
        return None
    
    def extract_class_members_with_clang(self, class_name, preprocessed_source):
        """Use Clang AST to extract ground truth class members."""
        members = {
            'methods': [],
            'fields': [],
            'static_fields': [],
            'constructors': [],
            'destructor': None,
            'all_members': []  # For debugging
        }
        
        # Parse the preprocessed source with Clang
        try:
            tu = self.clang_index.parse(
                'tmp.cpp',
                args=['-std=c++17', '-x', 'c++'],
                unsaved_files=[('tmp.cpp', preprocessed_source)]
            )
            
            # Check for parse errors
            if tu.diagnostics:
                severe_errors = [d for d in tu.diagnostics if d.severity >= clang.cindex.Diagnostic.Error]
                if severe_errors:
                    print(f"  {deco.yellow('Clang parse warnings/errors:')}")
                    for diag in severe_errors[:3]:  # Show first 3
                        print(f"    {diag.spelling}")
            
            # Find the class
            class_cursor = self.find_class_cursor(tu.cursor, class_name)
            if not class_cursor:
                print(f"  {deco.red('Class not found by Clang')}: {class_name}")
                return members
            
            # Track current access level (default is private for class, public for struct)
            current_access = AccessSpecifier.PRIVATE
            if class_cursor.kind == CursorKind.STRUCT_DECL:
                current_access = AccessSpecifier.PUBLIC
            
            # Traverse class members
            for child in class_cursor.get_children():
                # Update access specifier
                if child.kind == CursorKind.CXX_ACCESS_SPEC_DECL:
                    current_access = child.access_specifier
                    continue
                
                # Skip base specifiers and other non-member items
                if child.kind == CursorKind.CXX_BASE_SPECIFIER:
                    continue
                    
                # Constructors
                if child.kind == CursorKind.CONSTRUCTOR:
                    members['constructors'].append({
                        'signature': self.get_method_signature(child),
                        'access': self.get_access_string(current_access)
                    })
                    members['all_members'].append(f"Constructor: {child.spelling}")
                    
                # Destructor
                elif child.kind == CursorKind.DESTRUCTOR:
                    members['destructor'] = {
                        'signature': child.spelling,
                        'access': self.get_access_string(current_access)
                    }
                    members['all_members'].append(f"Destructor: {child.spelling}")
                    
                # Regular methods
                elif child.kind == CursorKind.CXX_METHOD:
                    method_info = {
                        'name': child.spelling,
                        'signature': self.get_method_signature(child),
                        'access': self.get_access_string(current_access),
                        'is_virtual': child.is_virtual_method(),
                        'is_static': child.is_static_method(),
                        'is_const': child.is_const_method()
                    }
                    members['methods'].append(method_info)
                    members['all_members'].append(f"Method: {child.spelling}")
                    
                # Function declarations (might be static methods)
                elif child.kind == CursorKind.FUNCTION_DECL:
                    method_info = {
                        'name': child.spelling,
                        'signature': self.get_method_signature(child),
                        'access': self.get_access_string(current_access),
                        'is_static': True
                    }
                    members['methods'].append(method_info)
                    members['all_members'].append(f"Static Function: {child.spelling}")
                    
                # Member fields
                elif child.kind == CursorKind.FIELD_DECL:
                    field_type = child.type.spelling
                    field_info = {
                        'name': child.spelling,
                        'type': field_type,
                        'access': self.get_access_string(current_access),
                        'is_static': False
                    }
                    
                    # Check if it's an array
                    if child.type.kind == clang.cindex.TypeKind.CONSTANTARRAY:
                        field_info['is_array'] = True
                        field_info['array_size'] = child.type.element_count
                        
                    members['fields'].append(field_info)
                    members['all_members'].append(f"Field: {child.spelling} ({field_type})")
                    
                # Static member variables
                elif child.kind == CursorKind.VAR_DECL:
                    field_info = {
                        'name': child.spelling,
                        'type': child.type.spelling,
                        'access': self.get_access_string(current_access),
                        'is_static': True
                    }
                    members['static_fields'].append(field_info)
                    members['all_members'].append(f"Static Field: {child.spelling}")
                    
                # Type aliases and typedefs within the class
                elif child.kind in [CursorKind.TYPEDEF_DECL, CursorKind.TYPE_ALIAS_DECL]:
                    members['all_members'].append(f"Typedef: {child.spelling}")
                    
        except Exception as e:
            self.errors.append(f"Clang parsing error for {class_name}: {str(e)}")
            print(f"  {deco.red('Clang parsing error')}: {str(e)}")
            
        return members
    
    def compare_class_members(self, entity, clang_members):
        """Compare database members with Clang-extracted members."""
        class_name = entity.short_name
        db_members = entity.members
        
        # Group database members by type
        db_methods = []
        db_fields = []
        db_static_fields = []
        db_constructors = []
        db_destructor = None
        
        for member in db_members:
            member_type = member.member_type.value
            if member_type == 'method':
                # Check if it's constructor/destructor by name
                if member.name == class_name:
                    db_constructors.append(member)
                elif member.name == f'~{class_name}':
                    db_destructor = member
                else:
                    db_methods.append(member)
            elif member_type == 'field':
                if member.is_static:
                    db_static_fields.append(member)
                else:
                    db_fields.append(member)
            elif member_type == 'static_field':
                db_static_fields.append(member)
        
        # Compare methods
        clang_method_names = {m['name'] for m in clang_members['methods']}
        db_method_names = {m.name for m in db_methods}
        
        # Check for missing methods
        missing_methods = clang_method_names - db_method_names
        for method_name in missing_methods:
            self.errors.append(f"Missing method in DB: {class_name}::{method_name}")
            self.stats['methods_missing'] += 1
            print(f"  {deco.red('Missing method')}: {method_name}")
            
        # Check for extra methods in DB
        extra_methods = db_method_names - clang_method_names
        for method_name in extra_methods:
            self.warnings.append(f"Extra method in DB: {class_name}::{method_name}")
            self.stats['methods_extra'] += 1
            print(f"  {deco.yellow('Extra method in DB')}: {method_name}")
            
        self.stats['methods_found'] += len(clang_method_names & db_method_names)
        
        # Check for overloads
        clang_method_sigs = defaultdict(list)
        for method in clang_members['methods']:
            clang_method_sigs[method['name']].append(method['signature'])
            
        db_method_sigs = defaultdict(list)
        for method in db_methods:
            db_method_sigs[method.name].append(method.signature or '')
            
        # Find overloaded methods
        for method_name, sigs in clang_method_sigs.items():
            if len(sigs) > 1:
                self.stats['overloads_detected'] += 1
                db_sigs = db_method_sigs.get(method_name, [])
                if len(db_sigs) != len(sigs):
                    self.errors.append(f"Overload count mismatch: {class_name}::{method_name} - Clang: {len(sigs)}, DB: {len(db_sigs)}")
                    print(f"  {deco.red('Overload mismatch')}: {method_name} - Clang: {len(sigs)}, DB: {len(db_sigs)}")
                else:
                    print(f"  {deco.cyan('Overloaded method')}: {method_name} ({len(sigs)} overloads)")
        
        # Compare fields
        clang_field_names = {f['name'] for f in clang_members['fields']}
        db_field_names = {f.name for f in db_fields}
        
        missing_fields = clang_field_names - db_field_names
        for field_name in missing_fields:
            # Find the field info
            field_info = next((f for f in clang_members['fields'] if f['name'] == field_name), None)
            if field_info:
                self.errors.append(f"Missing field in DB: {class_name}::{field_name} ({field_info['type']})")
                self.stats['fields_missing'] += 1
                print(f"  {deco.red('Missing field')}: {field_name} ({field_info['type']})")
                
        extra_fields = db_field_names - clang_field_names
        for field_name in extra_fields:
            self.warnings.append(f"Extra field in DB: {class_name}::{field_name}")
            self.stats['fields_extra'] += 1
            print(f"  {deco.yellow('Extra field in DB')}: {field_name}")
            
        self.stats['fields_found'] += len(clang_field_names & db_field_names)
        
        # Compare static fields
        clang_static_names = {f['name'] for f in clang_members['static_fields']}
        db_static_names = {f.name for f in db_static_fields}
        
        missing_static = clang_static_names - db_static_names
        for field_name in missing_static:
            field_info = next((f for f in clang_members['static_fields'] if f['name'] == field_name), None)
            if field_info:
                self.errors.append(f"Missing static field in DB: {class_name}::{field_name}")
                print(f"  {deco.red('Missing static field')}: {field_name}")
                
        # Check access levels for matching members
        for db_method in db_methods:
            clang_method = next((m for m in clang_members['methods'] if m['name'] == db_method.name), None)
            if clang_method:
                db_access = db_method.access_level.value
                clang_access = clang_method['access']
                if db_access != clang_access:
                    self.warnings.append(f"Access mismatch for {class_name}::{db_method.name}: DB={db_access}, Clang={clang_access}")
                    self.stats['access_mismatches'] += 1
    
    def validate_class(self, entity):
        """Validate a single class entity using Clang."""
        class_name = entity.short_name
        canonical_name = entity.canonical_name
        
        print(f"\n{deco.cyan('Checking class')}: {canonical_name}")
        self.stats['classes_checked'] += 1
        
        # Get locations for this class
        locations = entity.locations
        if not locations:
            self.warnings.append(f"No locations found for class: {class_name}")
            return
            
        # Get preprocessed source for the file containing class definition
        definition_location = None
        for loc in locations:
            if loc.location_type.value == 'definition':
                definition_location = loc
                break
        
        if not definition_location:
            # Use first location if no definition found
            definition_location = locations[0]
            
        # Get preprocessed source
        source_file = self.db.get_source_file(definition_location.file_path)
        if not source_file:
            self.warnings.append(f"No source file found for: {definition_location.file_path}")
            return
            
        preprocessed_source = source_file.get('preprocessed_source')
        if not preprocessed_source:
            # Fall back to raw source
            preprocessed_source = source_file.get('raw_source', '')
            print(f"  {deco.yellow('Warning')}: Using raw source instead of preprocessed")
            
        # Extract members using Clang
        clang_members = self.extract_class_members_with_clang(class_name, preprocessed_source)
        
        # Debug output
        if clang_members['all_members']:
            print(f"  {deco.green('Clang found')} {len(clang_members['all_members'])} members")
        
        # Compare with database
        self.compare_class_members(entity, clang_members)
    
    def run_validation(self, limit=None, stable=False):
        """Run validation on all classes in database."""
        print(f"{deco.green('=== Clang-Based Class Parsing Validation ===')}") 
        print(f"Database: {self.db.db_path}")
        
        # Get all class and struct entities
        classes = []
        for entity_type in ['class', 'struct']:
            entities = self.db.search_entities(
                entity_type=entity_type,
                limit=limit or 10000
            )
            classes.extend(entities)
        
        print(f"Found {len(classes)} classes/structs to validate")
        
        if limit:
            if stable:
                print(f"Limiting validation to first {limit} classes (stable selection)")
                classes = classes[:limit]
            else:
                import random
                print(f"Limiting validation to {limit} random classes for better coverage")
                classes = random.sample(classes, min(limit, len(classes)))
        
        # Validate each class
        for entity in classes:
            try:
                self.validate_class(entity)
            except Exception as e:
                self.errors.append(f"Error validating {entity.short_name}: {str(e)}")
                print(f"  {deco.red('Error')}: {str(e)}")
        
        # Print summary
        print(f"\n{deco.green('=== Validation Summary ===')}") 
        print(f"Classes checked: {self.stats['classes_checked']}")
        print(f"Methods found: {self.stats['methods_found']}")
        print(f"Methods missing: {self.stats['methods_missing']}")
        print(f"Methods extra: {self.stats['methods_extra']}")
        print(f"Fields found: {self.stats['fields_found']}")
        print(f"Fields missing: {self.stats['fields_missing']}")
        print(f"Fields extra: {self.stats['fields_extra']}")
        print(f"Overloaded methods detected: {self.stats['overloads_detected']}")
        print(f"Access level mismatches: {self.stats['access_mismatches']}")
        
        if self.errors:
            print(f"\n{deco.red(f'Errors ({len(self.errors)}):')}") 
            for error in self.errors[:10]:  # Show first 10 errors
                print(f"  - {error}")
            if len(self.errors) > 10:
                print(f"  ... and {len(self.errors) - 10} more")
                
        if self.warnings:
            print(f"\n{deco.yellow(f'Warnings ({len(self.warnings)}):')}") 
            for warning in self.warnings[:10]:  # Show first 10 warnings
                print(f"  - {warning}")
            if len(self.warnings) > 10:
                print(f"  ... and {len(self.warnings) - 10} more")
        
        # Write detailed report
        report_path = Path("/tmp/clang_validation_report.json")
        report = {
            'stats': self.stats,
            'errors': self.errors,
            'warnings': self.warnings
        }
        with open(report_path, 'w') as f:
            json.dump(report, f, indent=2)
        print(f"\nDetailed report written to: {report_path}")
        
        return len(self.errors) == 0

def main():
    import argparse
    
    parser = argparse.ArgumentParser(
        description='Validate C++ class parsing using Clang AST as ground truth'
    )
    parser.add_argument('--limit', type=int, 
                       help='Limit number of classes to check')
    parser.add_argument('--class', dest='class_name',
                       help='Check specific class only')
    parser.add_argument('--stable', action='store_true',
                       help='Use stable (non-random) class selection for reproducible results')
    
    args = parser.parse_args()
    
    # Get database path
    db_path = obt.path.stage() / "cpp_db_v2_orkid.db"
    
    if not db_path.exists():
        print(f"{deco.red(f'Database not found: {db_path}')}")
        print("Run ork.cpp.db.build.py first to build the database")
        sys.exit(1)
    
    # Run validation
    validator = ClangClassValidator(db_path)
    
    if args.class_name:
        # Validate specific class
        entities = validator.db.search_entities(name=args.class_name)
        if not entities:
            print(f"{deco.red(f'Class not found: {args.class_name}')}")
            sys.exit(1)
        
        for entity in entities:
            if entity.entity_type.value in ['class', 'struct']:
                validator.validate_class(entity)
    else:
        # Validate all classes
        success = validator.run_validation(limit=args.limit, stable=args.stable)
        sys.exit(0 if success else 1)

if __name__ == '__main__':
    main()


#!/usr/bin/env python3
###############################################################################
# Orkid Build System
# Copyright 2010-2018, Michael T. Mayers
# email: michael@tweakoz.com
# The Orkid Build System is published under the GPL 2.0 license
# see http://www.gnu.org/licenses/gpl-2.0.html
###############################################################################

import os, sys, string
import obt.search
import obt.search_args

#################################################################################

if __name__ == "__main__":
 args, ext_set = obt.search_args.parse_search_args(
     description='Search text in all OBT projects',
     require_dep=False
 )
 
 # execute expects a single word, not a list
 for word in args.keywords:
     obt.search.execute(word, ext_set=ext_set)



###############################################################################
# Orkid Build System
# Copyright 2010-2018, Michael T. Mayers
# email: michael@tweakoz.com
# The Orkid Build System is published under the GPL 2.0 license
# see http://www.gnu.org/licenses/gpl-2.0.html
###############################################################################

import os, tarfile, sys
from obt import dep, host, path, git, cmake, make
from obt.deco import Deco
from obt.wget import wget
from obt.command import Command

sys.path.append(os.path.dirname(os.path.abspath(__file__)))

import _gcc

deco = Deco()

###############################################################################

class irix65_gcc(dep.Provider):

    def __init__(self): ############################################
        super().__init__("irix65_gcc")
        self._archlist = ["x86_64"]


    ########

    def __str__(self):
        return "irix65 GCC (source)"

    ########

    def wipe(self):
        os.system("rm -rf %s"%self.source_root)
        os.system("rm -rf %s"%self.build_dest)

    ########

    def provide(self): ##########################################################
        if True: #False==self.OK:

            toolchain_dir = path.prefix()/"opt"/"toolchain"/"mips-sgi-irix6.5"
            self.toolchain_dir = toolchain_dir
            ###############################################
            ## fetch and extract irix 6.5.30 development sysroot
            ###############################################
            self.irixroot_fname = "irix-root.6.5.30.tar.bz2"
            self.irixroot_extract_dir = toolchain_dir/"irix6.5.30-sysroot"
            if self.irixroot_extract_dir.exists():
                os.system("chmod -R ugo+w %s"%self.irixroot_extract_dir)
                os.system("rm -rf %s"%self.irixroot_extract_dir)
            self.irixroot_arc = dep.downloadAndExtract(["http://mirror.larbob.org/compilertron/irix-root.6.5.30.tar.bz2"],
                                                        self.irixroot_fname,
                                                        "bz2",
                                                        "bb9a2d81729b23327207e6af30fd241f",
                                                        self.irixroot_extract_dir)
            ###############################################
            # patch gcc
            ###############################################
            gccpatchdir = path.patches()/"gcc"
            gcc_patches = [gccpatchdir/"gcc.sgifixes.patch",
                           gccpatchdir/"gcc.sgifixlibstdcpp01.patch",
                           gccpatchdir/"gcc.sgifixlibstdcpp02.patch"]
            newlib_patches = [gccpatchdir/"newlib.sgifixes.patch"]
            ###############################################
            # build it
            ###############################################
            binutils = dep.require("irix65_binutils")
            gcc = _gcc.context("mips-sgi-irix65",gccpatches=gcc_patches,newlibpatches=newlib_patches)
            bdest = gcc.build_dir/".build"
            pfx = path.prefix()
            os.mkdir(bdest)
            os.chdir(bdest)
            return gcc.build( target="mips-sgi-irix6.5",
                              variant="newlib",
                              program_prefix="mips-sgi-irix6.5-",
                              install_prefix=self.toolchain_dir,
                              with_ld=toolchain_dir/"bin"/"mips-sgi-irix6.5-ld",
                              with_as=toolchain_dir/"bin"/"mips-sgi-irix6.5-as",
                              with_build_sysroot=self.irixroot_extract_dir,
                              with_dwarf2=True
                              )

            """cmd = Command(['../configure',
                           '--prefix=%s'%toolchain_dir,
                           '--target=mips-sgi-irix6.5',
                           '--enable-languages=c,c++',
                           '--disable-nls',
                           '--disable-libssp',
                           #'--with-dwarf2'
                           #'--with-ld=%s'%(toolchain_dir/"bin"/"mips-sgi-irix6.5-ld"),
                           #'--with-as=%s'%(toolchain_dir/"bin"/"mips-sgi-irix6.5-as"),
                           #'--with-build-sysroot=%s'%self.irixroot_extract_dir,
                           '--disable-shared',
                           '--disable-threads',
                           '--disable-libgomp',
                           '--libdir=%s'%(toolchain_dir/"lib"/"mips-sgi-irix6.5-gcc"/gcc.version),
                           ])"""

            #cmd.exec()
            #make.exec()
            #make.exec("install",parallelism=0.0)

            #self.OK = True
            #self.manifest.touch()
            #return self.OK

"""
../libgcc/../gcc -I../../../../libgcc/../include  -DHAVE_CC_TLS  -o _gcov_dump.o -MT _gcov_dump.o -MD -MP -MF _gcov_dump.dep -DL_gcov_dump -c ../../../../libgcc/libgcov-interface.c
In file included from ../../../../libgcc/unwind-dw2.c:412:
./md-unwind-support.h:37:10: fatal error: signal.h: No such file or directory
37 | #include <signal.h>
|          ^~~~~~~~~~
compilation terminated.
"""

"""

projects/staging-obt/builds/gcc-mips-sgi-irix65/gcc-10.1.0/.build-stage2/./gcc/xgcc -B/projects/staging-obt/builds/gcc-mips-sgi-irix65/gcc-10.1.0/.build-stage2/./gcc/ -B//mips-sgi-irix6.5/bin/ -B//mips-sgi-irix6.5/lib/ -isystem //mips-sgi-irix6.5/include -isystem //mips-sgi-irix6.5/sys-include --sysroot=/projects/staging-obt/opt/toolchain/mips-sgi-irix6.5/irix6.5.30-sysroot   -g -O2 -mabi=64 -O2 -g -O2 -DIN_GCC  -DCROSS_DIRECTORY_STRUCTURE  -W -Wall -Wno-narrowing -Wwrite-strings -Wcast-qual -Wstrict-prototypes -Wmissing-prototypes -Wold-style-definition  -isystem ./include  -I. -I. -I../../.././gcc -I../../../../libgcc -I../../../../libgcc/. -I../../../../libgcc/../gcc -I../../../../libgcc/../include   -g0  -finhibit-size-directive -fno-inline -fno-exceptions -fno-zero-initialized-in-bss -fno-toplevel-reorder -fno-tree-vectorize -fbuilding-libgcc -fno-stack-protector  -Dinhibit_libc  -I. -I. -I../../.././gcc -I../../../../libgcc -I../../../../libgcc/. -I../../../../libgcc/../gcc -I../../../../libgcc/../include  -o irix-crtn.o -MT irix-crtn.o -MD -MP -MF irix-crtn.dep -c ../../../../libgcc/config/mips/irix-crtn.S


In file included from ../../../../libgcc/unwind-dw2.c:412:
./md-unwind-support.h:37:10: fatal error: signal.h: No such file or directory
37 | #include <signal.h>
"""


import os 
from obt._dep_build import BaseBuilder
from obt._dep_impl import require
from obt import pathtools, path, _globals
from obt import cmake, make
from obt.command import Command
import obt.host
from collections.abc import Callable

###############################################################################

class CMakeBuilder(BaseBuilder):
  ###########################################
  def __init__(self,
               name,
               static_libs=False,
               macos_defaults=True,
               install_prefix=None,
               src_dir_override=None,
               modules_paths=[],
               os_env=dict()):
    super().__init__(name)
    self._minimal = False 
    self._install_prefix = install_prefix
    self._src_dir_override = src_dir_override
    self._modules_paths = modules_paths
    self._os_env = os_env
    ##################################
    # ensure environment cmake present
    ##################################
    self._cmakeenv = {
      "CMAKE_BUILD_TYPE": "Release",
    }
    self._osenv = {
    }

    if not static_libs:
      self._cmakeenv["BUILD_SHARED_LIBS"]="ON"

    ##################################
    # default OSX stuff
    ##################################
    if obt.host.IsOsx and macos_defaults:
      sysroot_cmd = Command(["xcrun","--show-sdk-path"],do_log=False)
      sysroot = sysroot_cmd.capture().replace("\n","")

      if obt.host.IsAARCH64:
        self._cmakeenv.update({"CMAKE_HOST_SYSTEM_PROCESSOR":"arm64"})
      else:
        self._cmakeenv.update({"CMAKE_HOST_SYSTEM_PROCESSOR":"x86_64"})

      self._cmakeenv.update({
        "CMAKE_OSX_DEPLOYMENT_TARGET:STRING":"11",
        "CMAKE_OSX_SYSROOT:STRING":sysroot,
        "CMAKE_MACOSX_RPATH": "1",
        "CMAKE_INSTALL_RPATH": path.libs(),
        "CMAKE_SKIP_INSTALL_RPATH:BOOL":"NO",
        "CMAKE_SKIP_RPATH:BOOL":"NO",
        "CMAKE_INSTALL_NAME_DIR": "@executable_path/../lib"
      })

    ##################################
    self._parallelism = 0.0 if _globals.tryBoolOption("serial") else 1.0
    ##################################
    # implicit dependencies
    ##################################
    if name!="cmake":
      self._deps += ["cmake"]
  ###############################################
  @property 
  def install_prefix(self):
    return path.prefix() if (self._install_prefix==None) else self._install_prefix
  ###########################################
  def requires(self,deplist):
    self._deps += deplist
  ###########################################
  def setCmVar(self,key,value):
    self._cmakeenv[key] = value
  ###########################################
  def setCmVars(self,othdict):
    for k in othdict:
      self._cmakeenv[k] = othdict[k]
  ###########################################
  @property 
  def cmakeEnvAsString(self):
    return " ".join(self.cmakeEnvAsStringList)
  ###########################################
  @property 
  def cmakeEnvAsStringList(self):
    args = []
    for k in self._cmakeenv:
      v = self._cmakeenv[k]
      args += ["-D%s=%s"%(k,v)]
    return args
  ###########################################
  def build(self,srcdir,blddir,wrkdir,incremental=False):
    print("srcdir<%s>"%srcdir)
    print("blddir<%s>"%blddir)
    print("wrkdir<%s>"%wrkdir)
    print("srcovr<%s>"%self._src_dir_override)

    ok2build = require(self._deps)
    if not ok2build:
      return False

    environ_cached = os.environ.copy()

    os.environ.update(self._os_env)
    
    if incremental:
      pathtools.mkdir(blddir,clean=False)
      pathtools.chdir(wrkdir)
      cmake_ctx = cmake.context(root=srcdir,
                                env=self._cmakeenv,
                                osenv=self._osenv,
                                builddir=blddir,
                                working_dir=wrkdir,
                                sourcedir=self._src_dir_override,
                                install_prefix=self.install_prefix,
                                modules_paths = self._modules_paths)
      ok2build = cmake_ctx.exec()==0
    else:
      pathtools.mkdir(blddir,clean=True,parents=True)
      pathtools.chdir(wrkdir)
      cmake_ctx = cmake.context(root=srcdir,
                                env=self._cmakeenv,
                                sourcedir=self._src_dir_override,
                                osenv=self._osenv)
      ok2build = cmake_ctx.exec()==0

    if ok2build:
      OK = (make.exec(parallelism=self._parallelism)==0)
      if OK and self._onPostBuild!=None:
        self._onPostBuild()
        os.environ = environ_cached
      return OK
    os.environ = environ_cached
    return False
  ###########################################
  def install(self,blddir):
    pathtools.chdir(blddir)
    OK = (make.exec("install",parallelism=0.0)==0)
    if OK and self._onPostInstall!=None:
       self._onPostInstall()
    return OK

  ###########################################



###############################################################################
# Orkid Build System
# Copyright 2010-2018, Michael T. Mayers
# email: michael@tweakoz.com
# The Orkid Build System is published under the GPL 2.0 license
# see http://www.gnu.org/licenses/gpl-2.0.html
###############################################################################

from pathlib import Path
import hashlib
import obt.path
from obt.command import Command

###############################################################################

def check_hash(output_path,desired):
  if output_path.exists():
    the_md5 = hashlib.md5()
    the_md5.update(output_path.read_bytes())
    actual = the_md5.hexdigest()
    if actual==desired:
      return True
    else:
      print("path<%s>"%output_path)
      print("desired<%s>"%desired)
      print("actual<%s>"%actual)
  return False

###############################################################################
# caching download
###############################################################################

def wget(urls=[],
         output_name=None,
         md5val=None):

  assert(obt.path.downloads().exists())
  output_path = obt.path.downloads()/output_name
  hash_ok = False
  if md5val!=None:
    hash_ok = check_hash(output_path,md5val)
    if hash_ok:
      return output_path

  if False==hash_ok:
    dl_succeeded = False
    for url in urls:
      res = Command(["wget",
                     "-O",
                     output_path,
                     #"--show-progress",
                     url]).exec()

      if res==0:
        dl_succeeded = True 
        if md5val != None:
          hash_ok = check_hash(output_path,md5val)
          if False==hash_ok:
            return None
        return output_path

  return None

################################################################################
# batch downloads
################################################################################
def batch_wget(fset):
  for k in fset.keys():
    v = fset[k]
    wget(urls=[k],output_name=v[0],md5val=v[1])



#!/usr/bin/env python3
###############################################################################
# Enhanced display functions for search results with aligned output
###############################################################################

import os
from collections import defaultdict
import obt.deco

deco = obt.deco.Deco()

def collect_and_display_results(word_list, path_list, remove_root=None, ext_set=None,
                               case_insensitive=False, regex=False, whole_word=False,
                               invert_match=False, files_only=False,
                               before_context=0, after_context=0,
                               multiple_regex_and=False):
    """
    Collect all results first, then display with perfect alignment and grouped by file.
    """
    from obt.search import search_at_root, search_at_root_multiple_and, ignore_folder_keys, find_multiple_and
    from obt.search_v2 import find_with_context, find_with_context_multiple_and
    from obt.path import Path
    
    # Collect all results grouped by file
    all_results = defaultdict(list)  # filepath -> list of (line_num, line_text, is_match)
    max_path_len = 0
    MAX_FILENAME_LEN = 60  # Maximum filename length
    
    for path in path_list:
        if before_context > 0 or after_context > 0:
            # Context mode - use find_with_context
            for root, dirs, files in os.walk(str(path)):
                for f in files:
                    filepath = os.path.join(root, f)
                    spl = os.path.splitext(filepath)
                    ext = spl[1]
                    
                    # Check if should ignore
                    ignore = False
                    for item in ignore_folder_keys:
                        if item in spl[0]:
                            ignore = True
                            break
                    
                    if ignore or (ext_set and ext not in ext_set):
                        continue
                    
                    # Handle multiple regex AND mode differently
                    if multiple_regex_and and regex:
                        # Process all patterns together with AND logic
                        for context_result in find_with_context_multiple_and(word_list, filepath,
                                                                            case_insensitive, invert_match,
                                                                            before_context, after_context):
                            # Store the filepath for display
                            path_obj = Path(filepath)
                            display_path = path_obj.sanitized
                            if remove_root:
                                r_root = str(remove_root) + "/"
                                display_path = display_path.replace(r_root, "")
                            
                            # Truncate middle if too long
                            if len(display_path) > MAX_FILENAME_LEN:
                                # Keep start and end, truncate middle with "..."
                                keep_start = MAX_FILENAME_LEN // 2 - 2  # -2 for "..."
                                keep_end = MAX_FILENAME_LEN - keep_start - 3  # -3 for "..."
                                display_path = display_path[:keep_start] + "..." + display_path[-keep_end:]
                            
                            # Track max path length
                            max_path_len = max(max_path_len, len(display_path))
                            
                            # Add all lines from this context group
                            for line_num, line_text, is_match in context_result:
                                all_results[display_path].append((line_num, line_text, is_match))
                            
                            # Add separator marker between match groups in same file
                            if all_results[display_path]:
                                all_results[display_path].append((None, None, None))  # Separator
                    else:
                        # Original behavior: process each word separately (OR logic)
                        for word in word_list:
                            for context_result in find_with_context(word, filepath, 
                                                                   case_insensitive, regex, 
                                                                   whole_word, invert_match,
                                                                   before_context, after_context):
                                # Store the filepath for display
                                path_obj = Path(filepath)
                                display_path = path_obj.sanitized
                                if remove_root:
                                    r_root = str(remove_root) + "/"
                                    display_path = display_path.replace(r_root, "")
                                
                                # Truncate middle if too long
                                if len(display_path) > MAX_FILENAME_LEN:
                                    # Keep start and end, truncate middle with "..."
                                    keep_start = MAX_FILENAME_LEN // 2 - 2  # -2 for "..."
                                    keep_end = MAX_FILENAME_LEN - keep_start - 3  # -3 for "..."
                                    display_path = display_path[:keep_start] + "..." + display_path[-keep_end:]
                                
                                # Track max path length
                                max_path_len = max(max_path_len, len(display_path))
                                
                                # Add all lines from this context group
                                for line_num, line_text, is_match in context_result:
                                    all_results[display_path].append((line_num, line_text, is_match))
                                
                                # Add separator marker between match groups in same file
                                if all_results[display_path]:
                                    all_results[display_path].append((None, None, None))  # Separator
        else:
            # Non-context mode
            if multiple_regex_and and regex:
                # Multiple regex patterns with AND logic
                results = search_at_root_multiple_and(word_list, str(path), ext_set=ext_set,
                                                     case_insensitive=case_insensitive,
                                                     invert_match=invert_match,
                                                     files_only=files_only)
                for item in results:
                    path_obj = Path(item.path)
                    display_path = path_obj.sanitized
                    if remove_root:
                        r_root = str(remove_root) + "/"
                        display_path = display_path.replace(r_root, "")
                    
                    # Truncate middle if too long
                    if len(display_path) > MAX_FILENAME_LEN:
                        # Keep start and end, truncate middle with "..."
                        keep_start = MAX_FILENAME_LEN // 2 - 2  # -2 for "..."
                        keep_end = MAX_FILENAME_LEN - keep_start - 3  # -3 for "..."
                        display_path = display_path[:keep_start] + "..." + display_path[-keep_end:]
                    
                    max_path_len = max(max_path_len, len(display_path))
                    
                    if files_only and item.lineno == 0:
                        # Just track the file for files_only mode
                        all_results[display_path] = []
                    else:
                        all_results[display_path].append((item.lineno, item.text, True))
            else:
                # Original behavior: process each word separately (OR logic)
                for word in word_list:
                    results = search_at_root(word, str(path), ext_set=ext_set,
                                            case_insensitive=case_insensitive,
                                            regex=regex, whole_word=whole_word,
                                            invert_match=invert_match,
                                            files_only=files_only)
                    
                    for item in results:
                        path_obj = Path(item.path)
                        display_path = path_obj.sanitized
                        if remove_root:
                            r_root = str(remove_root) + "/"
                            display_path = display_path.replace(r_root, "")
                        
                        # Truncate middle if too long
                        if len(display_path) > MAX_FILENAME_LEN:
                            # Keep start and end, truncate middle with "..."
                            keep_start = MAX_FILENAME_LEN // 2 - 2  # -2 for "..."
                            keep_end = MAX_FILENAME_LEN - keep_start - 3  # -3 for "..."
                            display_path = display_path[:keep_start] + "..." + display_path[-keep_end:]
                        
                        max_path_len = max(max_path_len, len(display_path))
                        
                        if files_only and item.lineno == 0:
                            # Just track the file for files_only mode
                            all_results[display_path] = []
                        else:
                            all_results[display_path].append((item.lineno, item.text, True))
    
    # Now display all results with proper alignment
    if not all_results:
        return
    
    # Print header for each search path
    for idx, search_path in enumerate(path_list):
        if idx == 0 or all_results:  # Only print if we have results
            print("/////////////////////////////////////////////////////////////")
            print("// searching path : %s" % search_path)
            print("/////////")
            break  # Only show first path that has results
    
    # Display each file's results
    first_file = True
    for filepath in sorted(all_results.keys()):
        lines = all_results[filepath]
        
        if files_only and not lines:
            # Files-only mode: just print the filename
            print(deco.path(filepath))
            continue
        
        # Remove trailing None separators
        while lines and lines[-1][0] is None:
            lines.pop()
        
        # Print blank line between files (except first)
        if not first_file:
            print()
        first_file = False
        
        # Display the lines
        # First, find all match lines to determine tree structure
        match_indices = []
        for idx, (line_num, line_text, is_match) in enumerate(lines):
            if line_num is not None and is_match:
                match_indices.append(idx)
        
        first_match_in_file = True
        
        for idx, (line_num, line_text, is_match) in enumerate(lines):
            # Handle separator
            if line_num is None:
                if idx > 0 and idx < len(lines) - 1:  # Don't print separator at start/end
                    # Show vertical line during separators if more matches follow
                    if any(mi > idx for mi in match_indices):
                        path_text = deco.path(" " * (max_path_len - 3) + "│  ")
                        print(deco.formatColumns([(max_path_len + 2, 'right')], [path_text]))
                    else:
                        print()  # Just blank line
                continue
            
            # Determine what to show in the filename column
            if is_match:
                if first_match_in_file:
                    # First match in file: show the full filename
                    path_text = deco.path(filepath)
                    first_match_in_file = False
                else:
                    # Subsequent matches: show tree connector
                    # Check if this is the last match
                    is_last_match = (idx == match_indices[-1])
                    connector = "└──" if is_last_match else "├──"
                    # Position connector properly - vertical line on left, branch extends right
                    path_text = deco.path(" " * (max_path_len - 3) + connector)
            else:
                # Context line - show vertical line if between matches
                if match_indices and idx > match_indices[0] and idx < match_indices[-1]:
                    # We're between first and last match - vertical line on left
                    path_text = deco.path(" " * (max_path_len - 3) + "│  ")
                else:
                    path_text = ""
            
            # Format line number and text
            if is_match:
                line_num_text = deco.yellow(str(line_num))
                content_text = deco.val(line_text.strip())
            else:
                # Context lines
                line_num_text = deco.cyan(str(line_num))
                content_text = deco.inf(line_text.strip())
            
            # Use formatColumns for proper alignment
            # Columns: [filename, line number, content]
            column_configs = [
                (max_path_len + 2, 'right'),  # Filename column, right-aligned
                (8, 'right'),                  # Line number column, right-aligned
                (None, 'left')                 # Content column, no padding
            ]
            column_texts = [path_text, line_num_text, "  " + content_text]
            
            print(deco.formatColumns(column_configs, column_texts))


"""
Display detailed information about C++ classes and structs
"""
from typing import List, Dict, Set, Optional, Tuple
from pathlib import Path
import obt.deco as deco
import obt.path
from obt.cpp_database_v2 import CppDatabaseV2
from obt.cpp_entities_v2 import Entity, EntityType, Member, MemberType, AccessLevel, LocationType
from obt.cpp_formatter import CppFormatter
from obt.cpp_search_utils import find_entities_by_name, group_entities_by_namespace

class ClassDetailsDisplay:
    """Display detailed class/struct information"""
    
    def __init__(self, db: CppDatabaseV2):
        self.db = db
        self.deco = deco.Deco()
        
    def display_details(self, class_name: str, show_files: bool = True, root_path: Optional[Path] = None):
        """Display detailed information about a class/struct"""
        # Find all entities with this name (handles both short and canonical names)
        entities = find_entities_by_name(self.db, class_name)
        
        if not entities:
            print(f"{self.deco.red(f'Class/struct not found: {class_name}')}")
            return
        
        # Group by namespace to handle multiple classes with same name
        entities_by_ns = group_entities_by_namespace(entities)
        
        # Display each namespace separately if multiple
        for i, (namespace, namespace_entities) in enumerate(sorted(entities_by_ns.items())):
            if len(entities_by_ns) > 1:
                if i > 0:
                    print("\n" + "="*80 + "\n")
                print(f"{self.deco.cyan(f'NAMESPACE: {namespace}')}")
            
            # Use the first entity in this namespace (should be the most complete)
            entity = namespace_entities[0]
            self._display_single_class(entity, show_files, root_path)
    
    def _display_single_class(self, entity: Entity, show_files: bool, root_path: Optional[Path]):
        """Display details for a single class"""
        # Header
        primary_loc = entity.get_primary_location()
        file_path = None
        if primary_loc:
            file_path = str(obt.path.Path(primary_loc.file_path).sanitized)
        
        print(f"\n/////////////////////////////////////////////////////////////")
        print(f"// {entity.entity_type.value.title()} Details: {entity.short_name}")
        if file_path:
            print(f"// {file_path}")
        print(f"/////////\n")
        
        # Basic information
        self._display_basic_info(entity)
        
        # Base classes
        if entity.base_classes:
            self._display_base_classes(entity)
        
        # Template information
        if entity.is_template:
            self._display_template_info(entity)
        
        # Members by access level
        self._display_members(entity, show_files, root_path)
        
        # Method implementations
        if show_files:
            self._display_method_implementations(entity, root_path)
        
        # Locations
        if show_files and entity.locations:
            self._display_locations(entity, root_path)
    
    def _display_basic_info(self, entity: Entity):
        """Display basic class information"""
        print(f"{self.deco.yellow('BASIC INFO:')}")
        
        # Type and name
        type_str = entity.entity_type.value
        name_str = entity.canonical_name
        if entity.is_template and entity.template_params:
            name_str += entity.template_params
        print(f"  Type: {self.deco.cyan(type_str)} {self.deco.white(name_str)}")
        
        # Namespace
        ns_str = entity.namespace if entity.namespace else "(global)"
        print(f"  Namespace: {self.deco.green(ns_str)}")
        
        # Flags
        flags = []
        if entity.is_abstract:
            flags.append("abstract")
        if entity.is_final:
            flags.append("final")
        if entity.is_pod:
            flags.append("POD")
        if entity.is_template:
            flags.append("template")
        if entity.is_template_specialization:
            flags.append("specialization")
        
        if flags:
            print(f"  Attributes: {self.deco.magenta(', '.join(flags))}")
    
    def _display_base_classes(self, entity: Entity):
        """Display base class information"""
        print(f"\n{self.deco.yellow('BASE CLASSES:')}")
        for i, base in enumerate(entity.base_classes, 1):
            # Try to find the base class in the database
            base_clean = base.replace('public ', '').replace('private ', '').replace('protected ', '').strip()
            base_entities = []
            for entity_type in ['class', 'struct']:
                found = self.db.search_entities(entity_type=entity_type, name=base_clean)
                base_entities.extend(found)
            
            access_color = self.deco.green if 'public' in base else self.deco.red if 'private' in base else self.deco.yellow
            
            if base_entities:
                base_entity = base_entities[0]
                base_loc = base_entity.get_primary_location()
                if base_loc:
                    file_path = str(obt.path.Path(base_loc.file_path).sanitized)
                    print(f"  {i}. {access_color(base)} → {self.deco.cyan(file_path)}:{base_loc.line_number}")
                else:
                    print(f"  {i}. {access_color(base)} → {self.deco.white('(location unknown)')}")
            else:
                print(f"  {i}. {access_color(base)} → {self.deco.red('(not found in database)')}")
    
    def _display_template_info(self, entity: Entity):
        """Display template information"""
        print(f"\n{self.deco.yellow('TEMPLATE INFO:')}")
        if entity.template_params:
            print(f"  Parameters: {self.deco.cyan(entity.template_params)}")
        if entity.is_template_specialization and entity.specialized_from:
            print(f"  Specialized from: {self.deco.magenta(entity.specialized_from)}")
    
    def _display_members(self, entity: Entity, show_files: bool, root_path: Optional[Path]):
        """Display class members organized by access level"""
        if not entity.members:
            print(f"\n{self.deco.yellow('MEMBERS:')} {self.deco.gray('(none found)')}")
            return
        
        # Group members by access level
        public_members = entity.get_members_by_access(AccessLevel.PUBLIC)
        protected_members = entity.get_members_by_access(AccessLevel.PROTECTED)
        private_members = entity.get_members_by_access(AccessLevel.PRIVATE)
        
        # Display each access level
        for access_level, members, color_func in [
            ("PUBLIC", public_members, self.deco.green),
            ("PROTECTED", protected_members, self.deco.yellow),
            ("PRIVATE", private_members, self.deco.red)
        ]:
            if members:
                print(f"\n{self.deco.yellow(f'{access_level} MEMBERS:')} ({len(members)})")
                self._display_member_group(members, color_func, show_files, root_path, entity)
    
    def _display_member_group(self, members: List[Member], access_color_func, show_files: bool, root_path: Optional[Path], entity: Entity):
        """Display a group of members with the same access level"""
        # Group by member type
        fields = [m for m in members if m.member_type == MemberType.FIELD]
        methods = [m for m in members if m.member_type == MemberType.METHOD]
        constructors = [m for m in members if m.member_type == MemberType.CONSTRUCTOR]
        destructors = [m for m in members if m.member_type == MemberType.DESTRUCTOR]
        nested_types = [m for m in members if m.member_type == MemberType.NESTED_TYPE]
        enums = [m for m in members if m.member_type == MemberType.ENUM_VALUE]
        typedefs = [m for m in members if m.member_type == MemberType.TYPEDEF]
        
        # Display each type group
        for group_name, group_members in [
            ("Fields", fields),
            ("Constructors", constructors),
            ("Destructors", destructors),
            ("Methods", methods),
            ("Nested Types", nested_types),
            ("Enums", enums),
            ("Typedefs", typedefs)
        ]:
            if group_members:
                print(f"  {self.deco.cyan(group_name)}:")
                for member in sorted(group_members, key=lambda m: m.name):
                    self._display_member(member, access_color_func, show_files, root_path, entity)
    
    def _display_member(self, member: Member, access_color_func, show_files: bool, root_path: Optional[Path], entity: Entity):
        """Display a single member"""
        # Build member description
        parts = []
        
        # Handle method signatures with proper coloring
        if member.member_type == MemberType.METHOD and member.signature:
            # Apply basic coloring to the signature while preserving all content
            signature = member.signature
            
            # Simple approach: colorize based on basic patterns in the signature
            # This preserves all parameters while adding colors like function view
            colored_signature = self._colorize_method_signature(signature)
            parts.append(colored_signature)
            
            # Add additional modifiers that aren't part of the signature (like override, final, etc.)
            modifiers = []
            if member.is_override:
                modifiers.append("override")
            if member.is_final:
                modifiers.append("final")
            if member.is_deleted:
                modifiers.append("= delete")
            if member.is_default:
                modifiers.append("= default")
            
            if modifiers:
                parts.append(" " + self.deco.magenta(" ".join(modifiers)))
        else:
            # For non-methods, show modifiers first
            modifiers = []
            if member.is_static:
                modifiers.append("static")
            if member.is_const:
                modifiers.append("const")
            if hasattr(member, 'is_constexpr') and member.is_constexpr:
                modifiers.append("constexpr")
            if member.is_inline:
                modifiers.append("inline")
            if member.is_explicit:
                modifiers.append("explicit")
            
            if modifiers:
                parts.append(self.deco.magenta(" ".join(modifiers)))
            
        # Handle non-method types
        if member.member_type != MemberType.METHOD:
            if member.member_type == MemberType.ENUM_VALUE and member.value:
                name_display = f"{member.name} = {member.value}"
                parts.append(self.deco.white(name_display))
            elif member.member_type == MemberType.FIELD and member.value:
                # For fields with initialization values
                if member.data_type:
                    parts.append(self.deco.cyan(member.data_type))
                # Show field name with array dimensions and initialization value
                name_display = member.name
                if hasattr(member, 'array_dimensions') and member.array_dimensions:
                    name_display += member.array_dimensions
                parts.append(self.deco.white(name_display))
                parts.append(self.deco.white(" = "))
                parts.append(self.deco.green(member.value))
            else:
                # For fields and other members without initialization, show type and name
                if member.data_type:
                    parts.append(self.deco.cyan(member.data_type))
                
                # Show field name with array dimensions if present
                name_display = member.name
                if hasattr(member, 'array_dimensions') and member.array_dimensions:
                    name_display += member.array_dimensions
                parts.append(self.deco.white(name_display))
        
        # Line number if available
        if show_files and member.line_number > 0:
            parts.append(self.deco.gray(f":{member.line_number}"))
        
        print(f"    • {' '.join(parts)}")
    
    def _find_method_implementations(self, member: Member) -> List[object]:
        """Find implementation locations for a method from the member's own implementation_locations"""
        if not member.name or member.member_type != MemberType.METHOD:
            return []
        
        # Return the implementation locations stored in the member
        return member.implementation_locations
    
    def _colorize_signature(self, signature: str) -> str:
        """Apply colors to signature components like function view"""
        # Simple colorization - can be enhanced later
        # For now, just return the signature as-is since we're putting it in gray parentheses
        return signature
    
    def _colorize_method_signature(self, signature: str) -> str:
        """Apply colors to method signature to match function view"""
        import re
        
        # Pattern to match C++ function signatures
        # This regex captures: [modifiers] return_type function_name(parameters) [const/noexcept/etc]
        pattern = r'^((?:virtual\s+|static\s+|inline\s+|explicit\s+)*)(.*?)\s+(\w+)\s*(\([^)]*\))\s*(.*?)$'
        
        match = re.match(pattern, signature.strip())
        if not match:
            # Fallback: just color the whole thing white if we can't parse it
            return self.deco.white(signature)
        
        modifiers, return_type, func_name, params, trailing = match.groups()
        
        result_parts = []
        
        # Modifiers (virtual, static, etc.) - white like in function view
        if modifiers:
            result_parts.append(self.deco.white(modifiers))
        
        # Return type - magenta like in function view  
        if return_type:
            result_parts.append(self.deco.magenta(return_type.strip()))
            result_parts.append(self.deco.white(' '))
        
        # Function name - white
        result_parts.append(self.deco.white(func_name))
        
        # Parameters - color them like function view
        if params:
            colored_params = self._colorize_parameters(params)
            result_parts.append(colored_params)
        
        # Trailing modifiers (const, noexcept, etc.) - white
        if trailing:
            result_parts.append(self.deco.white(' ' + trailing.strip()))
        
        return ''.join(result_parts)
    
    def _colorize_parameters(self, params_str: str) -> str:
        """Color parameters like the function view: types in yellow, names in orange"""
        import re
        
        # Handle empty or void parameters
        if not params_str or params_str.strip() in ['()', '(void)']:
            return self.deco.white(params_str)
        
        # Extract content inside parentheses
        if params_str.startswith('(') and params_str.endswith(')'):
            inner = params_str[1:-1].strip()
            if not inner or inner == 'void':
                return self.deco.white(params_str)
            
            # Split parameters by comma
            param_parts = []
            current_param = []
            paren_depth = 0
            angle_depth = 0
            
            for char in inner:
                if char == '(' :
                    paren_depth += 1
                elif char == ')':
                    paren_depth -= 1
                elif char == '<':
                    angle_depth += 1
                elif char == '>':
                    angle_depth -= 1
                elif char == ',' and paren_depth == 0 and angle_depth == 0:
                    param_parts.append(''.join(current_param).strip())
                    current_param = []
                    continue
                current_param.append(char)
            
            if current_param:
                param_parts.append(''.join(current_param).strip())
            
            # Color each parameter
            colored_params = []
            for param in param_parts:
                colored_params.append(self._colorize_single_parameter(param.strip()))
            
            return self.deco.white('(') + ', '.join(colored_params) + self.deco.white(')')
        
        return self.deco.white(params_str)
    
    def _colorize_single_parameter(self, param: str) -> str:
        """Color a single parameter: type in yellow, name in orange"""
        import re
        
        # Pattern to match type and optional name
        # This handles cases like: "int x", "const char* name", "std::vector<int> vec", "Context"
        match = re.match(r'^(.*?)(\s+\w+)?$', param.strip())
        if not match:
            return self.deco.white(param)
        
        type_part = match.group(1).strip()
        name_part = match.group(2)
        
        result = []
        if type_part:
            result.append(self.deco.yellow(type_part))
        
        if name_part:
            result.append(self.deco.orange(name_part))
        
        return ''.join(result) if result else self.deco.white(param)
    
    def _display_method_implementations(self, entity: Entity, root_path: Optional[Path]):
        """Display method implementation locations in a separate section, grouped by signature"""
        # Collect all method implementations grouped by signature
        method_implementations = {}
        
        for member in entity.members:
            if member.member_type == MemberType.METHOD:
                impl_locations = self._find_method_implementations(member)
                if impl_locations:
                    # Use the full signature as the key to separate overloads
                    signature_key = member.signature if member.signature else f"{member.name}()"
                    method_key = f"{member.name}::{signature_key}"
                    
                    # Each unique signature gets its own entry
                    if method_key not in method_implementations:
                        method_implementations[method_key] = {
                            'name': member.name,
                            'signature': member.signature or f"{member.name}()",
                            'locations': impl_locations
                        }
                    else:
                        # If we have the same signature, add the locations
                        method_implementations[method_key]['locations'].extend(impl_locations)
        
        if not method_implementations:
            return
        
        print(f"\n{self.deco.yellow('METHOD IMPLEMENTATIONS:')}")
        
        for method_key, method_data in sorted(method_implementations.items(), key=lambda x: (x[1]['name'], x[1]['signature'])):
            method_name = method_data['name']
            signature = method_data['signature']
            impl_locations = method_data['locations']
            
            # Display method name and signature (using same colors as function view)
            if signature and signature != method_name:
                # Use the same colorized signature approach as the methods section
                colored_signature = self._colorize_method_signature(signature)
                print(f"  {colored_signature}:")
            else:
                print(f"  {self.deco.white(method_name)}:")
            
            # Display implementation locations (same color scheme as standard view)
            for impl_loc in impl_locations:
                file_path = str(obt.path.Path(impl_loc.file_path).sanitized)
                print(f"    {self.deco.fg('grey7')}{file_path}{self.deco.reset()}:{self.deco.fg('grey9')}{str(impl_loc.line_number)}{self.deco.reset()}")
    
    def _display_locations(self, entity: Entity, root_path: Optional[Path]):
        """Display all locations where this entity appears"""
        print(f"\n{self.deco.yellow('LOCATIONS:')}")
        
        # Group by location type
        definitions = [loc for loc in entity.locations if loc.location_type == LocationType.DEFINITION]
        declarations = [loc for loc in entity.locations if loc.location_type == LocationType.DECLARATION]
        forward_decls = [loc for loc in entity.locations if loc.location_type == LocationType.FORWARD_DECLARATION]
        
        for loc_type, locations, color_func in [
            ("Definitions", definitions, self.deco.green),
            ("Declarations", declarations, self.deco.cyan),
            ("Forward Declarations", forward_decls, self.deco.yellow)
        ]:
            if locations:
                print(f"  {color_func(loc_type)}:")
                for loc in sorted(locations, key=lambda l: (l.file_path, l.line_number)):
                    file_path = str(obt.path.Path(loc.file_path).sanitized)
                    body_indicator = " [has body]" if loc.has_body else ""
                    print(f"    {self.deco.white(file_path)}:{self.deco.yellow(str(loc.line_number))}{self.deco.gray(body_indicator)}")
                    if loc.context:
                        # Show a snippet of the context
                        context_lines = loc.context.strip().split('\n')[:3]  # Show first 3 lines
                        for line in context_lines:
                            print(f"      {self.deco.gray(line.strip())}")
                        if len(loc.context.strip().split('\n')) > 3:
                            print(f"      {self.deco.gray('...')}")


"""
C++ Entity Display Module for V2 Database
Standalone implementation for displaying C++ entities with formatting and colors
"""

from collections import defaultdict
from typing import List, Optional
from pathlib import Path
from obt.deco import Deco
from obt.cpp_formatter import CppFormatter
from obt.cpp_entities_v2 import Entity, EntityType, MemberType, AccessLevel
import obt.path
import json

class CppEntityDisplayV2:
    """Handles display of C++ entities with formatting and colors"""
    
    def __init__(self, line_length: int = 148):
        self.LINE_LENGTH = line_length
        self.deco = Deco()
        
        # Define column widths
        self.line_col_width = 8   # Line number
        self.type_col_width = 10  # Match type
        self.kind_col_width = 12  # Entity kind
        self.namespace_col_width = 30  # Namespace
        self.identifier_col_width = self.LINE_LENGTH - (
            self.line_col_width + self.type_col_width + 
            self.kind_col_width + self.namespace_col_width
        )  # Identifier gets the rest
    
    def get_match_type(self, entity: Entity) -> str:
        """Determine the match type for an entity"""
        # For V2, we primarily have definitions
        # Could extend to check location types if needed
        return 'DEF'
    
    def get_entity_kind(self, entity: Entity) -> str:
        """Get the entity kind for display"""
        if entity.entity_type == EntityType.FUNCTION:
            if entity.is_method:
                return "method"
            else:
                return "function"
        else:
            return entity.entity_type.value.lower()
    
    def _render_tagged_string(self, tagged_string) -> str:
        """Render a TaggedString with appropriate colors"""
        result = ""
        for text, tag in tagged_string.parts:
            if tag == 'keyword':
                result += self.deco.magenta(text)
            elif tag == 'type':
                result += self.deco.cyan(text)
            elif tag == 'param':
                result += self.deco.green(text)
            elif tag == 'default':
                result += self.deco.yellow(text)
            elif tag == 'template':
                result += self.deco.orange(text)
            elif tag == 'name':
                result += self.deco.white(text)
            else:
                result += self.deco.white(text)
        return result
    
    def get_kind_color(self, kind: str):
        """Get color function for entity kind"""
        kind_colors = {
            'class': self.deco.cyan,
            'struct': self.deco.cyan,
            'function': self.deco.cyan,
            'method': self.deco.cyan,
            'enum': self.deco.magenta,
            'typedef': self.deco.yellow,
            'alias': self.deco.yellow,
            'namespace': self.deco.blue,
            'variable': self.deco.green,
            'union': self.deco.orange,
        }
        return kind_colors.get(kind, self.deco.white)
    
    def display_entities(self, entities: List[Entity], root_path: Optional[Path] = None, 
                        sepfiles: bool = False, show_base_class: Optional[str] = None):
        """Display entities with full formatting
        
        Args:
            entities: List of entities to display
            root_path: Root path for relative display
            sepfiles: Add separator between files
            show_base_class: If set, show inheritance from this base class
        """
        if not entities:
            print("// No matching entities found")
            return
        
        # Group by file
        by_file = defaultdict(list)
        for entity in entities:
            if entity.locations:
                filepath = entity.locations[0].file_path
            else:
                filepath = "unknown"
            by_file[filepath].append(entity)
        
        print("/////////////////////////////////////////////////////////////")
        print(f"// Found {len(entities)} entities")
        print("/////////")
        
        file_count = 0
        
        for filepath in sorted(by_file.keys()):
            file_entities = by_file[filepath]
            
            # Add separator between files if requested
            if sepfiles and file_count > 0:
                print()
            
            # Print filename header (pink on grey1 background, centered)
            print()
            display_path = str(obt.path.Path(filepath).sanitized) if filepath != "unknown" else "unknown"
            filename_padded = display_path.center(self.LINE_LENGTH)
            print(self.deco.bg('grey1') + self.deco.fg('pink') + 
                  filename_padded + self.deco.reset())
            
            file_count += 1
            
            # Sort entities by line number
            sorted_entities = sorted(file_entities, 
                                    key=lambda x: x.locations[0].line_number if x.locations else 0)
            
            for entity in sorted_entities:
                # Build display components
                line = entity.locations[0].line_number if entity.locations else 0
                match_type = self.get_match_type(entity)
                kind = self.get_entity_kind(entity)
                namespace_text = entity.namespace if entity.namespace else "-"
                
                # Truncate namespace if too long
                if len(namespace_text) > self.namespace_col_width - 2:
                    keep_start = 8
                    keep_end = self.namespace_col_width - 4 - keep_start
                    namespace_text = namespace_text[:keep_start] + "..." + namespace_text[-keep_end:]
                
                # Build identifier
                if entity.entity_type == EntityType.FUNCTION:
                    formatter = CppFormatter(deco=None)
                    formatter.set_namespace_context(entity.namespace or "")
                    
                    # Create a compatible object for formatter
                    class FuncEntity:
                        pass
                    func = FuncEntity()
                    func.name = entity.short_name
                    func.return_type = entity.return_type or ""
                    func.is_template = entity.is_template
                    func.template_params = entity.template_params or ""
                    func.is_virtual = entity.is_virtual
                    func.is_static = entity.is_static
                    func.is_const = entity.is_const
                    func.is_method = entity.is_method
                    
                    # Convert parameters
                    if entity.parameters:
                        params_list = []
                        for p in entity.parameters:
                            param_dict = {
                                'type': p.param_type,
                                'name': p.name or '',
                                'default': p.default_value or ''
                            }
                            params_list.append(param_dict)
                        func.parameters = params_list
                    else:
                        func.parameters = []
                    
                    identifier = formatter.format_function_signature(func)
                else:
                    identifier = entity.short_name
                    if entity.is_template and entity.template_params:
                        identifier = f"template{entity.template_params} {identifier}"
                    
                    # Add inheritance info if requested
                    if show_base_class and entity.base_classes:
                        for base in entity.base_classes:
                            # Clean up base class name
                            base_clean = base.replace('public ', '').replace('private ', '').replace('protected ', '')
                            base_short = base_clean.split('::')[-1] if '::' in base_clean else base_clean
                            search_short = show_base_class.split('::')[-1] if '::' in show_base_class else show_base_class
                            if base_clean == show_base_class or base_short == search_short:
                                identifier += f" : {base}"
                                break
                
                # Format columns
                line_str = str(line).center(self.line_col_width)
                type_str = match_type.center(self.type_col_width)
                kind_str = kind.center(self.kind_col_width)
                namespace_str = namespace_text.center(self.namespace_col_width)
                
                # Get colors
                kind_color = self.get_kind_color(kind)
                
                # Handle the identifier display
                if hasattr(identifier, 'parts'):
                    # It's a TaggedString from CppFormatter - render it with colors
                    identifier_text = self._render_tagged_string(identifier)
                    plain_len = len(identifier.plain_text())
                    if plain_len < self.identifier_col_width:
                        padding = ' ' * (self.identifier_col_width - plain_len)
                        identifier_rendered = self.deco.bg('grey2') + identifier_text + padding + self.deco.reset()
                    else:
                        identifier_rendered = self.deco.bg('grey2') + identifier_text + self.deco.reset()
                else:
                    # Simple string identifier
                    padded_identifier = str(identifier).ljust(self.identifier_col_width) if len(str(identifier)) < self.identifier_col_width else str(identifier)
                    identifier_rendered = self.deco.bg('grey2') + self.deco.white(padded_identifier) + self.deco.reset()
                
                colored_columns = [
                    self.deco.bg('black') + self.deco.yellow(line_str) + self.deco.reset(),
                    self.deco.bg('grey2') + self.deco.green(type_str) + self.deco.reset(),
                    self.deco.bg('grey4') + kind_color(kind_str) + self.deco.reset(),
                    self.deco.bg('grey1') + self.deco.brightgreen(namespace_str) + self.deco.reset(),
                    identifier_rendered
                ]
                
                print(''.join(colored_columns))


#!/usr/bin/env python3

import zmq
import json
import time
import sys
import ssl
import socket

class SSLZMQClient:
    def __init__(self, host="localhost", ssl_port=8443, verify_cert=False):
        """
        Connect to ZMQ server through SSL proxy (stunnel)
        stunnel terminates SSL and forwards to ZMQ server
        """
        self.host = host
        self.ssl_port = ssl_port
        
        # Create SSL context
        if verify_cert:
            self.ssl_context = ssl.create_default_context()
        else:
            # For demo with self-signed certificates
            print("WARNING: SSL certificate verification is DISABLED")
            print("This is acceptable for demo purposes but NEVER in production!")
            self.ssl_context = ssl.create_default_context()
            self.ssl_context.check_hostname = False
            self.ssl_context.verify_mode = ssl.CERT_NONE
        
        # We'll create the SSL connection manually and use ZMQ over it
        self.socket = None
        self.connected = False
    
    def connect(self):
        """Establish SSL connection to stunnel proxy"""
        try:
            # Create TCP socket
            raw_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            
            # Wrap with SSL
            self.socket = self.ssl_context.wrap_socket(
                raw_socket, 
                server_hostname=self.host if self.ssl_context.check_hostname else None
            )
            
            # Connect to stunnel
            print(f"Connecting to SSL proxy at {self.host}:{self.ssl_port}")
            self.socket.connect((self.host, self.ssl_port))
            
            print(f"SSL connection established")
            print(f"Cipher: {self.socket.cipher()}")
            
            self.connected = True
            return True
            
        except Exception as e:
            print(f"Failed to connect: {e}")
            return False
    
    def send_request(self, command, data=None):
        """Send a ZMQ-style request over SSL connection"""
        if not self.connected:
            raise Exception("Not connected")
        
        # Prepare request
        request = {"command": command}
        if data is not None:
            request["data"] = data
        
        request_str = json.dumps(request)
        print(f"Sending: {request_str}")
        
        # Send request (ZMQ format over SSL)
        message = request_str.encode('utf-8')
        self.socket.send(message)
        
        # Receive response
        response_bytes = self.socket.recv(4096)
        response_str = response_bytes.decode('utf-8')
        print(f"Received: {response_str}")
        
        try:
            return json.loads(response_str)
        except json.JSONDecodeError:
            return {"error": "Invalid JSON response", "raw": response_str}
    
    def test_commands(self):
        """Test various server commands"""
        print("\n=== Testing SSL ZMQ Connection ===\n")
        
        if not self.connect():
            print("Failed to establish SSL connection")
            return
        
        try:
            # Test ping
            print("1. Testing ping command:")
            response = self.send_request("ping")
            print(f"   Status: {response.get('status')}")
            print(f"   Message: {response.get('message')}")
            print()
            
            # Test echo
            print("2. Testing echo command:")
            test_data = "Hello from SSL client!"
            response = self.send_request("echo", test_data)
            print(f"   Status: {response.get('status')}")
            print(f"   Echo: {response.get('message')}")
            print()
            
            # Test info
            print("3. Testing info command:")
            response = self.send_request("info")
            print(f"   Status: {response.get('status')}")
            print(f"   Server: {response.get('server')}")
            print(f"   Version: {response.get('version')}")
            print()
            
            # Test unknown command
            print("4. Testing unknown command:")
            response = self.send_request("unknown")
            print(f"   Status: {response.get('status')}")
            print(f"   Message: {response.get('message')}")
            print(f"   Available: {response.get('available_commands')}")
            print()
            
        except Exception as e:
            print(f"Error during communication: {e}")
    
    def cleanup(self):
        print("Cleaning up SSL client...")
        if self.socket:
            self.socket.close()
        self.connected = False

def main():
    if len(sys.argv) > 1:
        host = sys.argv[1]
    else:
        host = "localhost"
    
    client = SSLZMQClient(host=host, verify_cert=False)
    
    try:
        client.test_commands()
        print("=== SSL ZMQ Client Test Complete ===")
    except Exception as e:
        print(f"Error: {e}")
    finally:
        client.cleanup()

if __name__ == "__main__":
    main() 


#!/usr/bin/env python3
"""
Recursive Descent C++ Parser for tree-sitter AST
Based on comprehensive grammar analysis of 50 Orkid classes
Handles all node types properly including pure virtual methods
"""

from dataclasses import dataclass
from typing import Optional, List, Dict, Any
from pathlib import Path
import tree_sitter_cpp as ts_cpp
from tree_sitter import Language, Parser, Node

from .cpp_entities_v2 import (
    Entity, Member, Parameter, Location,
    EntityType, MemberType, AccessLevel, LocationType
)
from .cpp_type_system import TypeInfo, TypeRegistry, compose_type

class RecursiveDescentCppParser:
    """
    Clean recursive descent parser following tree-sitter-cpp grammar v0.22.3
    Properly handles all field_declaration_list child node types
    """
    
    def __init__(self, verbose: bool = False, db_path: Optional[str] = None):
        self.verbose = verbose
        self.parser = Parser()
        self.parser.set_language(Language(ts_cpp.language(), "cpp"))
        self.current_namespace = []
        self.current_file = None
        # Initialize type registry for unified type system
        self.type_registry = TypeRegistry(db_path) if db_path else None
        self.entities = []
        
    def parse_file(self, filepath: Path, source_code: Optional[str] = None) -> List[Entity]:
        """Parse a C++ file and return all entities"""
        self.current_file = filepath
        self.entities = []
        self.current_namespace = []
        
        # Read source if not provided
        if source_code is None:
            source_code = filepath.read_text(encoding='utf-8', errors='ignore')
        
        # Parse with tree-sitter
        tree = self.parser.parse(source_code.encode('utf-8'))
        source_bytes = source_code.encode('utf-8')
        
        # Start recursive descent
        self._parse_translation_unit(tree.root_node, source_bytes)
        
        return self.entities
    
    def parse_preprocessed(self, source_code: str, filepath: Path) -> List[Entity]:
        """Parse preprocessed source code"""
        return self.parse_file(filepath, source_code)
    
    # ============================================================
    # MAIN RECURSIVE DESCENT PARSER
    # ============================================================
    
    def _parse_translation_unit(self, node: Node, source: bytes):
        """Parse top-level translation unit"""
        for child in node.children:
            self._parse_top_level_item(child, source)
    
    def _parse_top_level_item(self, node: Node, source: bytes):
        """Parse any top-level declaration or definition"""
        if node.type == 'namespace_definition':
            self._parse_namespace_definition(node, source)
        elif node.type == 'class_specifier':
            entity = self._parse_class_specifier(node, source, is_struct=False)
            if entity:
                self.entities.append(entity)
        elif node.type == 'struct_specifier':
            entity = self._parse_class_specifier(node, source, is_struct=True)
            if entity:
                self.entities.append(entity)
        elif node.type == 'function_definition':
            entity = self._parse_function_definition_top_level(node, source)
            if entity:
                self.entities.append(entity)
        elif node.type == 'template_declaration':
            self._parse_template_declaration_top_level(node, source)
        elif node.type == 'enum_specifier':
            entity = self._parse_enum_specifier(node, source)
            if entity:
                self.entities.append(entity)
        elif node.type == 'type_definition':
            entity = self._parse_type_definition(node, source)
            if entity:
                self.entities.append(entity)
        elif node.type == 'alias_declaration':
            entity = self._parse_alias_declaration_top_level(node, source)
            if entity:
                self.entities.append(entity)
        elif node.type == 'declaration':
            # Could be variable, function declaration, etc.
            self._parse_declaration_top_level(node, source)
    
    # ============================================================
    # NAMESPACE HANDLING
    # ============================================================
    
    def _parse_namespace_definition(self, node: Node, source: bytes):
        """Parse namespace definition - handles both simple and nested namespaces"""
        namespace_parts = []
        
        for child in node.children:
            if child.type == 'namespace_identifier':
                # Simple namespace like 'namespace ork'
                namespace_parts.append(self._extract_text(child, source))
            elif child.type == 'nested_namespace_specifier':
                # Nested namespace like 'namespace ork::lev2'
                namespace_parts = self._extract_nested_namespace(child, source)
                break
        
        # Push all namespace parts
        for part in namespace_parts:
            self.current_namespace.append(part)
        
        # Parse namespace body
        for child in node.children:
            if child.type == 'declaration_list':
                for item in child.children:
                    if item.type not in ['{', '}']:
                        self._parse_top_level_item(item, source)
        
        # Pop all namespace parts
        for _ in namespace_parts:
            self.current_namespace.pop()
    
    def _extract_nested_namespace(self, node: Node, source: bytes) -> List[str]:
        """Extract all parts from a nested namespace specifier like ork::lev2::detail"""
        parts = []
        
        def extract_recursive(n: Node):
            """Recursively extract namespace parts from nested structure"""
            for child in n.children:
                if child.type == 'namespace_identifier':
                    parts.append(self._extract_text(child, source))
                elif child.type == 'nested_namespace_specifier':
                    # Recursively process nested namespace specifiers
                    extract_recursive(child)
        
        extract_recursive(node)
        
        if self.verbose and parts:
            print(f"DEBUG: Extracted nested namespace parts: {parts}")
        return parts
    
    # ============================================================
    # CLASS/STRUCT PARSING
    # ============================================================
    
    def _parse_class_specifier(self, node: Node, source: bytes, is_struct: bool) -> Optional[Entity]:
        """Parse class or struct definition"""
        entity_type = EntityType.STRUCT if is_struct else EntityType.CLASS
        namespace = '::'.join(self.current_namespace) if self.current_namespace else None
        
        has_body = False
        base_clause_node = None
        short_name = None
        
        # First pass: extract name, base classes, and check for body
        for child in node.children:
            if child.type == 'type_identifier':
                short_name = self._extract_text(child, source)
            elif child.type == 'base_class_clause':
                base_clause_node = child
            elif child.type == 'field_declaration_list':
                has_body = True
        
        if not short_name:
            return None
        
        # Now create entity with required fields
        full_name = self._build_full_name(short_name)
        entity = Entity(
            canonical_name=full_name,
            short_name=short_name,
            entity_type=entity_type
        )
        entity.namespace = namespace
        
        # Add location
        entity.locations.append(Location(
            file_path=str(self.current_file),
            line_number=source[:node.start_byte].count(b'\n') + 1,
            column_number=node.start_point[1] + 1,
            location_type=LocationType.DEFINITION if has_body else LocationType.DECLARATION
        ))
        
        # Parse base classes
        if base_clause_node:
            self._parse_base_class_clause(base_clause_node, source, entity)
        
        # Parse members if has body
        if has_body:
            for child in node.children:
                if child.type == 'field_declaration_list':
                    self._parse_field_declaration_list(
                        child, source, entity,
                        default_access=AccessLevel.PUBLIC if is_struct else AccessLevel.PRIVATE
                    )
        
        return entity
    
    def _parse_field_declaration_list(self, node: Node, source: bytes, entity: Entity,
                                     default_access: AccessLevel):
        """
        Parse class/struct body - THE CRITICAL METHOD
        Handles ALL member node types according to grammar
        """
        current_access = default_access
        
        for child in node.children:
            if child.type in ['{', '}', ';']:
                continue
                
            elif child.type == 'access_specifier':
                # Update access level
                current_access = self._parse_access_specifier(child, source)
                
            elif child.type == 'field_declaration':
                # Regular fields and method declarations
                member = self._parse_field_declaration(child, source, current_access)
                if member:
                    entity.add_member(member)
                    
            elif child.type == 'function_definition':
                # Methods with bodies, pure virtual methods, etc.
                member = self._parse_function_definition_member(child, source, current_access)
                if member:
                    entity.add_member(member)
                    
            elif child.type == 'template_declaration':
                # Template methods or nested template classes
                member = self._parse_template_declaration_member(child, source, current_access)
                if member:
                    entity.add_member(member)
                    
            elif child.type == 'declaration':
                # Forward declarations, nested types, etc.
                member = self._parse_declaration_member(child, source, current_access)
                if member:
                    entity.add_member(member)
                    
            elif child.type == 'alias_declaration':
                # Type aliases (using declarations)
                member = self._parse_alias_declaration_member(child, source, current_access)
                if member:
                    entity.add_member(member)
                    
            elif child.type == 'using_declaration':
                # Using declarations
                member = self._parse_using_declaration(child, source, current_access)
                if member:
                    entity.add_member(member)
                    
            elif child.type == 'friend_declaration':
                # Friend declarations
                pass  # Skip for now
                
            elif child.type == 'static_assert_declaration':
                # Static assertions
                pass  # Skip for now
                
            elif child.type == 'type_definition':
                # Typedef inside class
                member = self._parse_type_definition_member(child, source, current_access)
                if member:
                    entity.add_member(member)
                    
            elif child.type == ':':
                # Colon after access specifier
                continue
            else:
                if self.verbose:
                    print(f"Unknown member node type: {child.type}")
    
    # ============================================================
    # MEMBER PARSING
    # ============================================================
    
    def _parse_field_declaration(self, node: Node, source: bytes, 
                                access_level: AccessLevel) -> Optional[Member]:
        """Parse field_declaration - could be field or method declaration"""
        
        if self.verbose:
            print(f"DEBUG: _parse_field_declaration with {len(node.children)} children")
            for child in node.children:
                print(f"  Child: {child.type}")
        
        # Check if it's a function declaration
        has_function_declarator = False
        for child in node.children:
            if child.type == 'function_declarator':
                has_function_declarator = True
                break
            elif child.type == 'pointer_declarator':
                # Check if pointer_declarator contains function_declarator
                for ptr_child in child.children:
                    if ptr_child.type == 'function_declarator':
                        has_function_declarator = True
                        break
        
        if has_function_declarator:
            return self._parse_method_declaration(node, source, access_level)
        else:
            return self._parse_field_member(node, source, access_level)
    
    def _parse_function_definition_member(self, node: Node, source: bytes,
                                         access_level: AccessLevel) -> Optional[Member]:
        """
        Parse function_definition - CRITICAL FOR INLINE METHODS
        Uses unified type system for return type composition.
        """
        member = Member(name="", member_type=MemberType.METHOD)
        member.member_type = MemberType.METHOD
        member.access_level = access_level
        
        function_declarator = None
        return_type_info = None
        
        # Parse all children to extract complete information
        for child in node.children:
            if child.type == 'storage_class_specifier':
                # Check for static, inline, etc.
                for sc_child in child.children:
                    if sc_child.type == 'static':
                        member.is_static = True
                    elif sc_child.type == 'inline':
                        member.is_inline = True
                        
            elif child.type == 'virtual':
                member.is_virtual = True
                
            elif child.type == 'function_declarator':
                function_declarator = child
                
            elif child.type == 'pointer_declarator':
                # Function declarator might be inside pointer_declarator for pointer return types
                for ptr_child in child.children:
                    if ptr_child.type == 'function_declarator':
                        function_declarator = ptr_child
                        # Collect return type info using unified system
                        if not return_type_info:
                            return_type_info = self._collect_type_info(node, source)
                        
            elif child.type == 'reference_declarator':
                # Function declarator might be inside reference_declarator for reference return types
                for ref_child in child.children:
                    if ref_child.type == 'function_declarator':
                        function_declarator = ref_child
                        # Collect return type info using unified system
                        if not return_type_info:
                            return_type_info = self._collect_type_info(node, source)
                
            elif child.type == 'pure_virtual_clause':
                # THIS IS THE KEY FIX - detect pure virtual!
                member.is_pure_virtual = True
                
            elif child.type == 'default_method_clause':
                member.is_default = True
                
            elif child.type == 'delete_method_clause':
                member.is_deleted = True
                    
            elif child.type == 'compound_statement':
                # Has implementation body
                pass
        
        # Collect return type if not already done
        if not return_type_info:
            return_type_info = self._collect_return_type_info(node, source)
        
        # Compose and store the return type using unified system
        if return_type_info and return_type_info.base_type:
            member.data_type = compose_type(return_type_info)
            # Store in flyweight registry if available
            if self.type_registry:
                member.base_type_id = self.type_registry.get_or_create_type(return_type_info)
                # Store type modifiers separately
                member.is_reference = return_type_info.is_reference
                member.is_rvalue_reference = return_type_info.is_rvalue_reference
                member.is_volatile = return_type_info.is_volatile
                member.return_const = return_type_info.is_const
        
        # Extract name and parameters from function_declarator
        if function_declarator:
            self._parse_function_declarator(function_declarator, source, member)
        
        # Determine if it's constructor/destructor
        if member.name:
            if member.name.startswith('~'):
                member.member_type = MemberType.DESTRUCTOR
            elif member.name == self._get_current_class_name():
                member.member_type = MemberType.CONSTRUCTOR
        
        # Set line number
        member.line_number = source[:node.start_byte].count(b'\n') + 1
        
        return member if member.name else None
    
    def _parse_method_declaration(self, node: Node, source: bytes,
                                 access_level: AccessLevel) -> Optional[Member]:
        """Parse method declaration from field_declaration with function_declarator
        Uses unified type system for return type composition.
        """
        member = Member(name="", member_type=MemberType.METHOD)
        member.member_type = MemberType.METHOD
        member.access_level = access_level
        
        function_declarator = None
        return_type_info = None
        
        if self.verbose:
            print(f"DEBUG: Parsing method declaration, node has {len(node.children)} children")
        
        for child in node.children:
            if self.verbose:
                print(f"  Child: {child.type}")
            
            if child.type == 'storage_class_specifier':
                for sc_child in child.children:
                    if sc_child.type == 'static':
                        member.is_static = True
                        if self.verbose:
                            print(f"    Found static keyword")
                    elif sc_child.type == 'inline':
                        member.is_inline = True
                        
            elif child.type == 'static':
                # Static can also appear directly
                member.is_static = True
                if self.verbose:
                    print(f"    Found static keyword (direct)")
                        
            elif child.type == 'virtual':
                member.is_virtual = True
                if self.verbose:
                    print(f"    Found virtual keyword")
                
            elif child.type == 'function_declarator':
                function_declarator = child
                
            elif child.type == 'pointer_declarator':
                # Function declarator might be inside pointer_declarator for pointer return types
                for ptr_child in child.children:
                    if ptr_child.type == 'function_declarator':
                        function_declarator = ptr_child
                        # Collect return type using unified system
                        if not return_type_info:
                            return_type_info = self._collect_type_info(node, source)
                
            elif child.type == 'pure_virtual_clause':
                member.is_pure_virtual = True
                if self.verbose:
                    print(f"    Found pure virtual clause (= 0)")
        
        # Collect return type if not already done
        if not return_type_info:
            return_type_info = self._collect_return_type_info(node, source)
        
        # Compose and store the return type using unified system
        if return_type_info and return_type_info.base_type:
            member.data_type = compose_type(return_type_info)
            # Store in flyweight registry if available
            if self.type_registry:
                member.base_type_id = self.type_registry.get_or_create_type(return_type_info)
                # Store type modifiers separately
                member.is_reference = return_type_info.is_reference
                member.is_rvalue_reference = return_type_info.is_rvalue_reference
                member.is_volatile = return_type_info.is_volatile
                member.return_const = return_type_info.is_const
        
        if function_declarator:
            self._parse_function_declarator(function_declarator, source, member)
            if self.verbose:
                print(f"  After parsing function_declarator: name='{member.name}', is_static={member.is_static}")
        
        # Check for constructor/destructor
        if member.name:
            if member.name.startswith('~'):
                member.member_type = MemberType.DESTRUCTOR
            elif member.name == self._get_current_class_name():
                member.member_type = MemberType.CONSTRUCTOR
        
        member.line_number = source[:node.start_byte].count(b'\n') + 1
        
        return member if member.name else None
    
    def _parse_field_member(self, node: Node, source: bytes,
                          access_level: AccessLevel) -> Optional[Member]:
        """Parse a field member using unified type system"""
        member = Member(name="", member_type=MemberType.FIELD)
        member.access_level = access_level
        
        # Check for nested class/struct definition first
        for child in node.children:
            if child.type in ['class_specifier', 'struct_specifier']:
                # This is a nested type definition
                type_name = self._extract_nested_type_name(child, source)
                if type_name:
                    member.name = type_name
                    member.member_type = MemberType.NESTED_TYPE
                    member.data_type = child.type.replace('_specifier', '')
                    member.line_number = source[:node.start_byte].count(b'\n') + 1
                    return member
        
        # Use unified type collection for ALL type information
        type_info = self._collect_type_info(node, source)
        
        # Extract field name and handle special declarators
        field_name = None
        field_value = None
        
        for child in node.children:
            if child.type == 'field_identifier':
                field_name = self._extract_text(child, source)
                
            elif child.type == 'init_declarator':
                # Has an initializer, extract name and value
                for init_child in child.children:
                    if init_child.type == 'field_identifier':
                        field_name = self._extract_text(init_child, source)
                    elif init_child.type == 'pointer_declarator':
                        # Field name is nested in pointer declarator
                        for ptr_child in init_child.children:
                            if ptr_child.type == 'field_identifier':
                                field_name = self._extract_text(ptr_child, source)
                    elif init_child.type == 'array_declarator':
                        # Field name is nested in array declarator
                        for arr_child in init_child.children:
                            if arr_child.type == 'field_identifier':
                                field_name = self._extract_text(arr_child, source)
                    elif init_child.type == 'reference_declarator':
                        # Field name is nested in reference declarator
                        for ref_child in init_child.children:
                            if ref_child.type == 'field_identifier':
                                field_name = self._extract_text(ref_child, source)
                    elif init_child.type == '=':
                        # Next sibling is the value
                        idx = list(child.children).index(init_child)
                        if idx + 1 < len(child.children):
                            field_value = self._extract_text(child.children[idx + 1], source)
                            
            elif child.type == 'pointer_declarator':
                # Extract name from pointer declarator
                for ptr_child in child.children:
                    if ptr_child.type == 'field_identifier':
                        field_name = self._extract_text(ptr_child, source)
                        
            elif child.type == 'array_declarator':
                # Extract name from array declarator
                for arr_child in child.children:
                    if arr_child.type == 'field_identifier':
                        field_name = self._extract_text(arr_child, source)
                        
            elif child.type == 'reference_declarator':
                # Extract name from reference declarator
                for ref_child in child.children:
                    if ref_child.type == 'field_identifier':
                        field_name = self._extract_text(ref_child, source)
                        
            elif child.type == '=' and not field_value:
                # Direct initializer at field level
                idx = list(node.children).index(child)
                if idx + 1 < len(node.children):
                    field_value = self._extract_text(node.children[idx + 1], source)
        
        # Store collected type information in member
        member.name = field_name
        member.value = field_value
        
        # Apply type modifiers from TypeInfo
        member.is_static = type_info.is_constexpr or member.is_static  # constexpr implies static
        member.is_const = type_info.is_const
        member.is_volatile = type_info.is_volatile
        member.is_constexpr = type_info.is_constexpr
        member.is_mutable = type_info.is_mutable
        member.is_reference = type_info.is_reference
        member.is_rvalue_reference = type_info.is_rvalue_reference
        member.pointer_depth = type_info.pointer_depth
        member.array_dimensions = ''.join(f'[{d}]' for d in type_info.array_dimensions) if type_info.array_dimensions else None
        
        # Compose and store the complete type string
        member.data_type = compose_type(type_info)
        
        # Store type in flyweight registry if available
        if self.type_registry and type_info.base_type:
            member.base_type_id = self.type_registry.get_or_create_type(type_info)
        
        member.line_number = source[:node.start_byte].count(b'\n') + 1
        
        return member if member.name else None
    
    def _parse_function_declarator(self, node: Node, source: bytes, member: Member):
        """Parse function_declarator to extract name, parameters, qualifiers"""
        if self.verbose:
            print(f"DEBUG: Parsing function_declarator with {len(node.children)} children")
            for child in node.children:
                print(f"  Child type: {child.type}")
        
        params = []
        
        # First pass: collect all information before building signature
        for child in node.children:
            if child.type in ['field_identifier', 'identifier', 'destructor_name', 'operator_name']:
                member.name = self._extract_text(child, source)
                if self.verbose:
                    print(f"  Found function name: {member.name}")
                
            elif child.type == 'parameter_list':
                params = self._parse_parameter_list(child, source)
                    
            elif child.type == 'type_qualifier':
                qual = self._extract_text(child, source)
                if qual == 'const':
                    member.is_const = True
                elif qual == 'noexcept':
                    member.is_noexcept = True
                    
            elif child.type == 'virtual_specifier':
                spec = self._extract_text(child, source)
                if spec == 'override':
                    member.is_override = True
                elif spec == 'final':
                    member.is_final = True
                    
            elif child.type == 'ref_qualifier':
                # Handle && and & qualifiers
                pass
        
        # Build signature with all qualifiers for proper overload detection
        if member.name and params is not None:
            param_strs = []
            for param in params:
                param_str = param.param_type
                if param.name:
                    param_str += f" {param.name}"
                if param.default_value:
                    param_str += f" = {param.default_value}"
                param_strs.append(param_str)

            # Build qualifiers string
            qualifiers = []
            if member.is_const:
                qualifiers.append("const")
            if member.is_noexcept:
                qualifiers.append("noexcept")
            if member.is_override:
                qualifiers.append("override")
            if member.is_final:
                qualifiers.append("final")
                
            qualifier_str = " ".join(qualifiers)
            if qualifier_str:
                qualifier_str = " " + qualifier_str

            if member.data_type:
                member.signature = f"{member.data_type} {member.name}({', '.join(param_strs)}){qualifier_str}"
            else:
                member.signature = f"{member.name}({', '.join(param_strs)}){qualifier_str}"
    
    def _parse_parameter_list(self, node: Node, source: bytes) -> List[Parameter]:
        """Parse parameter list"""
        params = []
        
        for child in node.children:
            if child.type == 'parameter_declaration':
                param = self._parse_parameter_declaration(child, source)
                if param:
                    params.append(param)
            elif child.type == 'optional_parameter_declaration':
                param = self._parse_optional_parameter_declaration(child, source)
                if param:
                    params.append(param)
            elif child.type == 'variadic_parameter_declaration':
                param = self._parse_variadic_parameter_declaration(child, source)
                if param:
                    params.append(param)
        
        return params
    
    def _parse_parameter_declaration(self, node: Node, source: bytes) -> Optional[Parameter]:
        """Parse a parameter declaration"""
        param_type = ""
        param_name = None
        is_reference = False
        is_pointer = False
        
        for child in node.children:
            if child.type in ['primitive_type', 'type_identifier', 'qualified_identifier']:
                param_type = self._extract_text(child, source)
            elif child.type == 'identifier':
                param_name = self._extract_text(child, source)
            elif child.type == 'reference_declarator':
                # Handle reference parameters
                is_reference = True
                for ref_child in child.children:
                    if ref_child.type == 'identifier':
                        param_name = self._extract_text(ref_child, source)
            elif child.type == 'pointer_declarator':
                # Handle pointer parameters
                is_pointer = True
                for ptr_child in child.children:
                    if ptr_child.type == 'identifier':
                        param_name = self._extract_text(ptr_child, source)
        
        if param_type:
            param = Parameter(name=param_name, param_type=param_type)
            param.is_reference = is_reference
            param.is_pointer = is_pointer
            return param
        return None
    
    def _parse_optional_parameter_declaration(self, node: Node, source: bytes) -> Optional[Parameter]:
        """Parse optional parameter (with default value)"""
        param = self._parse_parameter_declaration(node, source)
        if param:
            # Find default value
            for child in node.children:
                if child.type == '=':
                    # Next sibling should be the default value
                    idx = node.children.index(child)
                    if idx + 1 < len(node.children):
                        param.default_value = self._extract_text(node.children[idx + 1], source)
                    break
        return param
    
    def _parse_variadic_parameter_declaration(self, node: Node, source: bytes) -> Optional[Parameter]:
        """Parse variadic parameter (...)"""
        param = Parameter(name=None, param_type="...")
        return param
    
    def _parse_template_declaration_member(self, node: Node, source: bytes,
                                          access_level: AccessLevel) -> Optional[Member]:
        """Parse template declaration inside class"""
        # Find the actual declaration inside template
        for child in node.children:
            if child.type == 'function_definition':
                member = self._parse_function_definition_member(child, source, access_level)
                if member:
                    member.is_template = True
                return member
            elif child.type == 'declaration':
                member = self._parse_declaration_member(child, source, access_level)
                if member:
                    member.is_template = True
                return member
            elif child.type in ['class_specifier', 'struct_specifier']:
                # Nested template class
                type_name = self._extract_nested_type_name(child, source)
                if type_name:
                    member = Member(name="", member_type=MemberType.METHOD)
                    member.name = type_name
                    member.member_type = MemberType.NESTED_TYPE
                    member.data_type = f"template {child.type.replace('_specifier', '')}"
                    member.access_level = access_level
                    member.is_template = True
                    member.line_number = source[:node.start_byte].count(b'\n') + 1
                    return member
        return None
    
    def _parse_declaration_member(self, node: Node, source: bytes,
                                 access_level: AccessLevel) -> Optional[Member]:
        """Parse declaration inside class"""
        # Could be function declaration, variable declaration, etc.
        has_function = any(child.type == 'function_declarator' for child in node.children)
        
        if has_function:
            return self._parse_method_declaration(node, source, access_level)
        else:
            return self._parse_field_member(node, source, access_level)
    
    def _parse_alias_declaration_member(self, node: Node, source: bytes,
                                       access_level: AccessLevel) -> Optional[Member]:
        """Parse type alias (using declaration) inside class"""
        member = Member(name="", member_type=MemberType.METHOD)
        member.member_type = MemberType.NESTED_TYPE
        member.access_level = access_level
        
        alias_name = None
        alias_type = None
        
        for child in node.children:
            if child.type == 'type_identifier' and not alias_name:
                alias_name = self._extract_text(child, source)
            elif child.type in ['type_identifier', 'qualified_identifier', 'template_type']:
                if alias_name:  # This is the aliased type
                    alias_type = self._extract_text(child, source)
        
        if alias_name:
            member.name = alias_name
            member.data_type = f"using {alias_name} = {alias_type or 'unknown'}"
            member.line_number = source[:node.start_byte].count(b'\n') + 1
            return member
        
        return None
    
    def _parse_type_definition_member(self, node: Node, source: bytes,
                                     access_level: AccessLevel) -> Optional[Member]:
        """Parse typedef inside class"""
        member = Member(name="", member_type=MemberType.METHOD)
        member.member_type = MemberType.NESTED_TYPE
        member.access_level = access_level
        
        typedef_name = None
        typedef_type = None
        
        for child in node.children:
            if child.type == 'type_identifier':
                if not typedef_type:
                    typedef_type = self._extract_text(child, source)
                else:
                    typedef_name = self._extract_text(child, source)
        
        if typedef_name:
            member.name = typedef_name
            member.data_type = f"typedef {typedef_type or 'unknown'} {typedef_name}"
            member.line_number = source[:node.start_byte].count(b'\n') + 1
            return member
        
        return None
    
    def _parse_using_declaration(self, node: Node, source: bytes,
                                access_level: AccessLevel) -> Optional[Member]:
        """Parse using declaration inside class"""
        # Skip for now - rarely used for members
        return None
    
    # ============================================================
    # ACCESS SPECIFIER PARSING
    # ============================================================
    
    def _parse_access_specifier(self, node: Node, source: bytes) -> AccessLevel:
        """Parse access specifier and return access level"""
        for child in node.children:
            if child.type in ['public', 'private', 'protected']:
                text = self._extract_text(child, source)
                if text == 'public':
                    return AccessLevel.PUBLIC
                elif text == 'private':
                    return AccessLevel.PRIVATE
                elif text == 'protected':
                    return AccessLevel.PROTECTED
        return AccessLevel.PRIVATE  # Default
    
    # ============================================================
    # TOP-LEVEL ENTITY PARSING
    # ============================================================
    
    def _parse_function_definition_top_level(self, node: Node, source: bytes) -> Optional[Entity]:
        """Parse top-level function definition"""
        namespace = '::'.join(self.current_namespace) if self.current_namespace else None
        
        function_declarator = None
        return_type = None
        short_name = None
        
        for child in node.children:
            if child.type == 'function_declarator':
                function_declarator = child
            elif child.type in ['primitive_type', 'type_identifier', 'qualified_identifier', 'auto']:
                if not return_type:
                    return_type = self._extract_text(child, source)
        
        if function_declarator:
            # Extract function name
            for child in function_declarator.children:
                if child.type in ['identifier', 'qualified_identifier', 'operator_name']:
                    short_name = self._extract_text(child, source)
                    break
        
        if short_name:
            # Now create entity with required fields
            full_name = self._build_full_name(short_name)
            entity = Entity(
                canonical_name=full_name,
                short_name=short_name,
                entity_type=EntityType.FUNCTION
            )
            entity.namespace = namespace
            
            entity.locations.append(Location(
                file_path=str(self.current_file),
                line_number=source[:node.start_byte].count(b'\n') + 1,
                column_number=node.start_point[1] + 1,
                location_type=LocationType.DEFINITION
            ))
            return entity
        
        return None
    
    def _parse_template_declaration_top_level(self, node: Node, source: bytes):
        """Parse top-level template declaration"""
        # Find what's being templated
        for child in node.children:
            if child.type in ['class_specifier', 'struct_specifier']:
                entity = self._parse_class_specifier(
                    child, source, 
                    is_struct=(child.type == 'struct_specifier')
                )
                if entity:
                    entity.is_template = True
                    self.entities.append(entity)
            elif child.type == 'function_definition':
                entity = self._parse_function_definition_top_level(child, source)
                if entity:
                    entity.is_template = True
                    self.entities.append(entity)
    
    def _parse_enum_specifier(self, node: Node, source: bytes) -> Optional[Entity]:
        """Parse enum definition"""
        namespace = '::'.join(self.current_namespace) if self.current_namespace else None
        short_name = None
        
        for child in node.children:
            if child.type == 'type_identifier':
                short_name = self._extract_text(child, source)
                break
        
        if short_name:
            # Now create entity with required fields
            full_name = self._build_full_name(short_name)
            entity = Entity(
                canonical_name=full_name,
                short_name=short_name,
                entity_type=EntityType.ENUM
            )
            entity.namespace = namespace
            
            entity.locations.append(Location(
                file_path=str(self.current_file),
                line_number=source[:node.start_byte].count(b'\n') + 1,
                column_number=node.start_point[1] + 1,
                location_type=LocationType.DEFINITION
            ))
            return entity
        
        return None
    
    def _parse_type_definition(self, node: Node, source: bytes) -> Optional[Entity]:
        """Parse typedef"""
        namespace = '::'.join(self.current_namespace) if self.current_namespace else None
        
        # Find typedef name (usually the last type_identifier)
        type_identifiers = []
        for child in node.children:
            if child.type == 'type_identifier':
                type_identifiers.append(self._extract_text(child, source))
        
        if type_identifiers:
            short_name = type_identifiers[-1]
            full_name = self._build_full_name(short_name)
            
            # Now create entity with required fields
            entity = Entity(
                canonical_name=full_name,
                short_name=short_name,
                entity_type=EntityType.TYPEDEF
            )
            entity.namespace = namespace
            
            entity.locations.append(Location(
                file_path=str(self.current_file),
                line_number=source[:node.start_byte].count(b'\n') + 1,
                column_number=node.start_point[1] + 1,
                location_type=LocationType.DEFINITION
            ))
            return entity
        
        return None
    
    def _parse_alias_declaration_top_level(self, node: Node, source: bytes) -> Optional[Entity]:
        """Parse top-level type alias"""
        namespace = '::'.join(self.current_namespace) if self.current_namespace else None
        short_name = None
        
        for child in node.children:
            if child.type == 'type_identifier':
                short_name = self._extract_text(child, source)
                break
        
        if short_name:
            # Now create entity with required fields
            full_name = self._build_full_name(short_name)
            entity = Entity(
                canonical_name=full_name,
                short_name=short_name,
                entity_type=EntityType.TYPEDEF  # Treat as typedef
            )
            entity.namespace = namespace
            
            entity.locations.append(Location(
                file_path=str(self.current_file),
                line_number=source[:node.start_byte].count(b'\n') + 1,
                column_number=node.start_point[1] + 1,
                location_type=LocationType.DEFINITION
            ))
            return entity
        
        return None
    
    def _parse_declaration_top_level(self, node: Node, source: bytes):
        """Parse top-level declaration"""
        # Could be function declaration, variable, etc.
        has_function = any(child.type == 'function_declarator' for child in node.children)
        
        if has_function:
            # Function declaration
            namespace = '::'.join(self.current_namespace) if self.current_namespace else None
            short_name = None
            
            for child in node.children:
                if child.type == 'function_declarator':
                    for fc in child.children:
                        if fc.type in ['identifier', 'qualified_identifier']:
                            short_name = self._extract_text(fc, source)
                            break
            
            if short_name:
                # Now create entity with required fields
                full_name = self._build_full_name(short_name)
                entity = Entity(
                    canonical_name=full_name,
                    short_name=short_name,
                    entity_type=EntityType.FUNCTION
                )
                entity.namespace = namespace
                
                entity.locations.append(Location(
                    file_path=str(self.current_file),
                    line_number=source[:node.start_byte].count(b'\n') + 1,
                    column_number=node.start_point[1] + 1,
                    location_type=LocationType.DECLARATION
                ))
                self.entities.append(entity)
    
    def _parse_base_class_clause(self, node: Node, source: bytes, entity: Entity):
        """Parse base class clause for inheritance"""
        for child in node.children:
            if child.type == 'type_identifier':
                base_name = self._extract_text(child, source)
                entity.base_classes.append(base_name)
            elif child.type == 'qualified_identifier':
                base_name = self._extract_text(child, source)
                entity.base_classes.append(base_name)
    
    def _parse_reference_declarator(self, node: Node, source: bytes, member: Member):
        """Parse reference declarator - NEW METHOD"""
        for child in node.children:
            if child.type == 'field_identifier':
                member.name = self._extract_text(child, source)
            elif child.type == 'identifier':
                member.name = self._extract_text(child, source)
        
        # Add reference indication to data type
        if member.data_type:
            member.data_type += '&'
        else:
            member.data_type = '&'  # In case type wasn't set yet
    
    def _parse_pointer_declarator(self, node: Node, source: bytes, member: Member):
        """Parse pointer declarator"""
        for child in node.children:
            if child.type == 'field_identifier':
                member.name = self._extract_text(child, source)
            elif child.type == 'identifier':
                member.name = self._extract_text(child, source)
        
        # Add * to type
        if member.data_type:
            member.data_type += '*'
    
    def _parse_array_declarator(self, node: Node, source: bytes, member: Member):
        """Parse array declarator - FIXED for multi-dimensional arrays"""
        dimensions = []
        
        def extract_dimensions_recursive(n: Node):
            """Recursively extract dimensions from nested array_declarator nodes"""
            for child in n.children:
                if child.type == 'field_identifier':
                    member.name = self._extract_text(child, source)
                elif child.type == 'identifier':
                    # Could be field name OR array dimension - check context
                    text = self._extract_text(child, source)
                    if not member.name:
                        member.name = text
                    else:
                        # This is an array dimension identifier (e.g., [kAABUFTILES])
                        dimensions.append(text)
                elif child.type == 'number_literal':
                    dimensions.append(self._extract_text(child, source))
                elif child.type == 'array_declarator':
                    # Recursively handle nested array declarators
                    extract_dimensions_recursive(child)
        
        extract_dimensions_recursive(node)
        
        if dimensions:
            member.array_dimensions = ''.join(f'[{d}]' for d in dimensions)
    
    def _parse_init_declarator(self, node: Node, source: bytes, member: Member):
        """Parse init_declarator to extract field name and initializer value"""
        for child in node.children:
            if child.type == 'field_identifier':
                member.name = self._extract_text(child, source)
            elif child.type == 'pointer_declarator':
                self._parse_pointer_declarator(child, source, member)
            elif child.type == 'array_declarator':
                self._parse_array_declarator(child, source, member)  
            elif child.type == 'reference_declarator':
                self._parse_reference_declarator(child, source, member)
            elif child.type == '=':
                # Found initializer - extract the value
                # Find the next sibling that contains the initializer value
                found_equals = False
                for sibling in node.children:
                    if found_equals and sibling.type not in [';', ' ', '\n', '\t']:
                        member.value = self._extract_text(sibling, source)
                        break
                    if sibling == child:
                        found_equals = True
    
    # ============================================================
    # HELPER METHODS
    # ============================================================
    
    def _extract_text(self, node: Node, source: bytes) -> str:
        """Extract text from a node"""
        return source[node.start_byte:node.end_byte].decode('utf-8', errors='ignore')
    
    # ============================================================
    # UNIFIED TYPE SYSTEM METHODS
    # ============================================================
    
    def _collect_type_info(self, node: Node, source: bytes) -> TypeInfo:
        """
        Collect complete type information from AST nodes.
        Central method for ALL type extraction - replaces scattered type composition.
        """
        type_info = TypeInfo()
        
        # Process the node and its children to collect all type metadata
        self._collect_type_info_recursive(node, source, type_info)
        
        return type_info
    
    def _collect_return_type_info(self, node: Node, source: bytes) -> TypeInfo:
        """
        Collect return type information specifically for methods.
        This is more targeted than _collect_type_info to avoid mixing
        return type qualifiers with method qualifiers.
        """
        type_info = TypeInfo()
        
        # Look for return type nodes specifically
        for child in node.children:
            if child.type in ['primitive_type', 'type_identifier', 'qualified_identifier', 
                            'auto', 'template_type', 'sized_type_specifier']:
                # This is the base return type
                type_info.base_type = self._extract_text(child, source)
                if child.type == 'template_type':
                    type_info.is_template = True
                    
                # Check for const qualifier BEFORE the type
                idx = list(node.children).index(child)
                if idx > 0:
                    prev = node.children[idx - 1]
                    if prev.type == 'type_qualifier':
                        qual = self._extract_text(prev, source)
                        if qual == 'const':
                            type_info.is_const = True
                        elif qual == 'volatile':
                            type_info.is_volatile = True
                            
            elif child.type == 'pointer_declarator':
                # Check if function is inside pointer declarator (pointer return type)
                has_function = False
                for ptr_child in child.children:
                    if ptr_child.type == 'function_declarator':
                        has_function = True
                        break
                        
                if has_function:
                    # This is a pointer return type
                    type_info.pointer_depth = 1
                    # Get base type from before pointer_declarator
                    idx = list(node.children).index(child)
                    if idx > 0:
                        prev = node.children[idx - 1]
                        if prev.type in ['primitive_type', 'type_identifier', 'qualified_identifier', 'template_type']:
                            type_info.base_type = self._extract_text(prev, source)
                            if prev.type == 'template_type':
                                type_info.is_template = True
                                
                        # Check for const before type
                        if idx > 1:
                            prev2 = node.children[idx - 2]
                            if prev2.type == 'type_qualifier' and self._extract_text(prev2, source) == 'const':
                                type_info.is_const = True
                                
            elif child.type == 'reference_declarator':
                # Check if function is inside reference_declarator (reference return type)
                has_function = False
                for ref_child in child.children:
                    if ref_child.type == 'function_declarator':
                        has_function = True
                        break
                    elif ref_child.type == '&&':
                        type_info.is_rvalue_reference = True
                        
                if has_function:
                    # This is a reference return type
                    if not type_info.is_rvalue_reference:
                        type_info.is_reference = True
                        
                    # Get base type from before reference_declarator
                    idx = list(node.children).index(child)
                    if idx > 0:
                        prev = node.children[idx - 1]
                        if prev.type in ['primitive_type', 'type_identifier', 'qualified_identifier', 'template_type']:
                            type_info.base_type = self._extract_text(prev, source)
                            if prev.type == 'template_type':
                                type_info.is_template = True
                                
                        # Check for const before type
                        if idx > 1:
                            prev2 = node.children[idx - 2]
                            if prev2.type == 'type_qualifier' and self._extract_text(prev2, source) == 'const':
                                type_info.is_const = True
        
        return type_info
    
    def _collect_type_info_recursive(self, node: Node, source: bytes, type_info: TypeInfo):
        """Recursively collect type information from AST nodes"""
        
        # Handle type qualifiers
        if node.type == 'type_qualifier':
            qual = self._extract_text(node, source)
            if qual == 'const':
                type_info.is_const = True
            elif qual == 'volatile':
                type_info.is_volatile = True
            elif qual == 'mutable':
                type_info.is_mutable = True
                
        # Handle storage class specifiers
        elif node.type == 'storage_class_specifier':
            for child in node.children:
                spec = self._extract_text(child, source)
                if spec == 'constexpr':
                    type_info.is_constexpr = True
                elif spec == 'mutable':
                    type_info.is_mutable = True
                    
        # Handle base types
        elif node.type in ['primitive_type', 'type_identifier', 'qualified_identifier', 'auto']:
            if not type_info.base_type:
                type_info.base_type = self._extract_text(node, source)
                
        # Handle template types
        elif node.type == 'template_type':
            if not type_info.base_type:
                type_info.base_type = self._extract_text(node, source)
                type_info.is_template = True
                
        # Handle sized type specifiers (e.g., "unsigned int")
        elif node.type == 'sized_type_specifier':
            if not type_info.base_type:
                type_info.base_type = self._extract_text(node, source)
                
        # Handle pointer declarators
        elif node.type == 'pointer_declarator':
            type_info.pointer_depth += 1
            # Continue collecting from children
            for child in node.children:
                if child.type != '*':
                    self._collect_type_info_recursive(child, source, type_info)
                    
        # Handle reference declarators
        elif node.type == 'reference_declarator':
            type_info.is_reference = True
            # Continue collecting from children
            for child in node.children:
                if child.type not in ['&', '&&']:
                    self._collect_type_info_recursive(child, source, type_info)
                elif child.type == '&&':
                    type_info.is_rvalue_reference = True
                    type_info.is_reference = False  # rvalue ref is distinct from lvalue ref
                    
        # Handle array declarators
        elif node.type == 'array_declarator':
            # Extract array dimensions
            for child in node.children:
                if child.type == 'number_literal':
                    type_info.array_dimensions.append(self._extract_text(child, source))
                elif child.type == 'identifier':
                    # Could be array dimension constant
                    text = self._extract_text(child, source)
                    if text and text[0].isupper():  # Likely a constant
                        type_info.array_dimensions.append(text)
                elif child.type not in ['[', ']', 'field_identifier']:
                    self._collect_type_info_recursive(child, source, type_info)
                    
        # Recurse for other node types
        else:
            for child in node.children:
                self._collect_type_info_recursive(child, source, type_info)
    
    def _build_full_name(self, short_name: str) -> str:
        """Build full qualified name"""
        if self.current_namespace:
            return '::'.join(self.current_namespace) + '::' + short_name
        return short_name
    
    def _get_current_class_name(self) -> Optional[str]:
        """Get the name of the class currently being parsed"""
        # This would need to be tracked during parsing
        # For now, return None
        return None
    
    def _extract_nested_type_name(self, node: Node, source: bytes) -> Optional[str]:
        """Extract name from nested class/struct"""
        for child in node.children:
            if child.type == 'type_identifier':
                return self._extract_text(child, source)
        return None
    
    def parse_source(self, source_code: bytes, filepath: str) -> List[Entity]:
        """Parse source code bytes and return entities - compatible with database usage"""
        self.current_file = Path(filepath)
        self.entities = []
        self.current_namespace = []
        
        # Parse with tree-sitter
        tree = self.parser.parse(source_code)
        
        # Start recursive descent
        self._parse_translation_unit(tree.root_node, source_code)
        
        return self.entities

# Export alias for compatibility
CppParser = RecursiveDescentCppParser


"""
C++ Source File Ingestion Module
Reads raw files, preprocesses them, and stores in database
Designed for parallel execution
"""

import subprocess
import time
from pathlib import Path
from typing import List, Optional, Dict, Set
from concurrent.futures import ProcessPoolExecutor, as_completed
import multiprocessing
import obt.deco

class CppIngestor:
    """Handles ingestion of C++ source files into database"""
    
    def __init__(self, 
                 defines: Optional[Dict[str, str]] = None,
                 include_paths: Optional[List[str]] = None,
                 defines_preset: Optional[str] = None):
        self.defines = defines or {}
        self.include_paths = include_paths or []
        self.defines_preset = defines_preset
        self.deco = obt.deco.Deco()
        
    def preprocess_file(self, filepath: Path) -> tuple[Path, bytes, bytes, bytes, float]:
        """
        Preprocess a single file and trim to target file content only
        Returns: (filepath, raw_source, preprocessed_source, trimmed_source, preprocess_time)
        """
        import time
        start_time = time.perf_counter()
        
        # Read raw source
        with open(filepath, 'rb') as f:
            raw_source = f.read()
            
        # Build preprocessor command
        cmd = ['clang', '-E', '-x', 'c++', '-std=c++20']
        
        # Add defines
        for key, value in self.defines.items():
            if value:
                cmd.append(f'-D{key}={value}')
            else:
                cmd.append(f'-D{key}')
                
        # Add include paths
        for inc_path in self.include_paths:
            cmd.append(f'-I{inc_path}')
            
        # Add the file
        cmd.append(str(filepath))
        
        try:
            # Run preprocessor
            result = subprocess.run(cmd, capture_output=True, timeout=30)
            
            if result.stdout:
                preprocessed = result.stdout
            else:
                # Fallback to raw if preprocessing fails
                preprocessed = raw_source
                
        except subprocess.TimeoutExpired:
            print(f"Warning: Preprocessing timeout for {filepath}")
            preprocessed = raw_source
        except Exception as e:
            print(f"Warning: Preprocessing error for {filepath}: {e}")
            preprocessed = raw_source
            
        # Trim to target file content only
        trimmed = self._trim_to_target_file(preprocessed, filepath)
        
        preprocess_time = time.perf_counter() - start_time
            
        return filepath, raw_source, preprocessed, trimmed, preprocess_time
    
    def _trim_to_target_file(self, source_code: bytes, target_file: Path) -> bytes:
        """
        Remove all content from included files, keeping only the target file's content.
        This dramatically reduces the amount of code to parse.
        """
        import re
        
        # Normalize target file path for comparison
        target_file_str = str(target_file.resolve())
        target_file_name = target_file.name
        
        # Parse line directives: # linenum "filename" flags
        line_directive_pattern = re.compile(rb'^#\s+(\d+)\s+"([^"]+)".*?$', re.MULTILINE)
        
        result = []
        lines = source_code.split(b'\n')
        current_file = target_file_str
        keep_lines = True
        
        for line in lines:
            match = line_directive_pattern.match(line)
            if match:
                # This is a line directive
                filename = match.group(2).decode('utf-8', errors='ignore')
                
                # Check if we're entering or leaving the target file
                if filename == target_file_str or filename.endswith('/' + target_file_name) or filename == target_file_name:
                    keep_lines = True
                else:
                    keep_lines = False
                
                # Skip the directive itself
                continue
            
            # Only keep lines from the target file
            if keep_lines:
                result.append(line)
        
        return b'\n'.join(result)
    
    def ingest_files_parallel(self, filepaths: List[Path], 
                            max_workers: Optional[int] = None) -> Dict:
        """
        Ingest multiple files in parallel
        Returns dict of results
        """
        if max_workers is None:
            max_workers = multiprocessing.cpu_count()
            
        results = {
            'success': [],
            'failed': [],
            'stats': {
                'total_raw_bytes': 0,
                'total_preprocessed_bytes': 0,
                'time_seconds': 0
            }
        }
        
        start_time = time.time()
        
        with ProcessPoolExecutor(max_workers=max_workers) as executor:
            # Submit all files for processing
            futures = {
                executor.submit(self.preprocess_file, fp): fp 
                for fp in filepaths
            }
            
            # Process results as they complete
            completed = 0
            total = len(futures)
            
            for future in as_completed(futures):
                filepath = futures[future]
                try:
                    path, raw, preprocessed, trimmed, preprocess_time = future.result()
                    results['success'].append({
                        'path': path,
                        'raw_size': len(raw),
                        'preprocessed_size': len(preprocessed),
                        'trimmed_size': len(trimmed),
                        'raw_source': raw,
                        'preprocessed_source': preprocessed,
                        'trimmed_source': trimmed,
                        'preprocess_time': preprocess_time
                    })
                    results['stats']['total_raw_bytes'] += len(raw)
                    results['stats']['total_preprocessed_bytes'] += len(preprocessed)
                    
                except Exception as e:
                    results['failed'].append({
                        'path': filepath,
                        'error': str(e)
                    })
                
                completed += 1
                if completed % 10 == 0 or completed == total:
                    percent = (completed / total) * 100
                    deco = obt.deco.Deco()
                    # Orange for "Ingestion progress", white for percent, yellow for counter
                    progress_text = f"{deco.orange('Ingestion progress:')} {deco.white(f'{percent:.1f}%')} {deco.yellow(f'({completed}/{total})')}"
                    print(f"\r{progress_text}", end='', flush=True)
                    
        # Print newline after completion
        if completed == total:
            print()
            
        results['stats']['time_seconds'] = time.time() - start_time
        return results


def ingest_to_database(files: List[Path], db_path: Path, 
                       max_workers: Optional[int] = None) -> None:
    """
    Ingest files and store in database
    """
    from obt.cpp_database_v2 import CppDatabaseV2
    
    # Create ingestor
    ingestor = CppIngestor()
    
    # Process files in parallel
    print(f"Ingesting {len(files)} files with {max_workers or multiprocessing.cpu_count()} workers...")
    results = ingestor.ingest_files_parallel(files, max_workers)
    
    # Store in database
    db = CppDatabaseV2(db_path)
    
    with db.connect() as conn:
        for item in results['success']:
            # Store source file
            conn.execute("""
                INSERT OR REPLACE INTO source_files 
                (file_path, file_name, raw_source, preprocessed_source, file_size)
                VALUES (?, ?, ?, ?, ?)
            """, (
                str(item['path']),
                item['path'].name,
                item['raw_source'].decode('utf-8', errors='ignore'),
                item['preprocessed_source'].decode('utf-8', errors='ignore'),
                item['raw_size']
            ))
        conn.commit()
    
    # Print stats
    print(f"\nIngestion complete:")
    print(f"  Success: {len(results['success'])} files")
    print(f"  Failed: {len(results['failed'])} files")
    print(f"  Raw size: {results['stats']['total_raw_bytes']:,} bytes")
    print(f"  Preprocessed size: {results['stats']['total_preprocessed_bytes']:,} bytes")
    print(f"  Time: {results['stats']['time_seconds']:.2f} seconds")
    print(f"  Rate: {len(results['success']) / results['stats']['time_seconds']:.1f} files/second")


# Module exports
__all__ = ['CppIngestor', 'ingest_to_database']


from obt import command

class Session(object):
  #########################################
  def __init__(self, session_name, orientation="horizontal",working_dir=None,kill_first=True):
    self.session_name = session_name
    self.orientation = orientation
    self.working_dir = working_dir
    self.kill_first = kill_first
    self.cmd_new_session = [
      "tmux", "new-session",
      "-d", "-s", f"{self.session_name}"]
    if self.orientation == "vertical":
      self.cmd_add_session = [
        "tmux", "split-window",
        "-v", "-t", f"{self.session_name}"]
      self.cmd_select_layout = [
        "tmux", "select-layout",
        "-t", f"{self.session_name}",
        "even-vertical"
      ]
    elif self.orientation == "horizontal":
      self.cmd_add_session = [
        "tmux", "split-window",
        "-h", "-t", f"{self.session_name}"]
      self.cmd_select_layout = [
        "tmux", "select-layout",
        "-t", f"{self.session_name}",
        "even-horizontal"
      ]
    self.cmd_attach_session = [
      "tmux", "attach-session",
      "-t", f"{self.session_name}"
    ]
    self.kill_session = [
      "tmux", "kill-session",
      "-t", f"{self.session_name}"
    ]
    if working_dir!=None:
      self.cmd_new_session.append("-c")
      self.cmd_new_session.append(working_dir)
      self.cmd_add_session.append("-c")
      self.cmd_add_session.append(working_dir)

    self.cmd_chain = command.chain2(do_log=True)
    self.post_chain = list()
    self.num_panes = 1  # Start with 1 since new-session creates the first pane
    
    ######################
    # default key bindings
    ######################
    self.bind_key("K", "kill-session", table="prefix")
    self.bind_key("H", "select-layout even-horizontal", table="prefix")
    self.bind_key("V", "select-layout even-vertical", table="prefix")

  #########################################
  def bind_zoom_panes(self):
    """Bind Shift+1 through Shift+num_panes to zoom respective panes"""
    # Map shift+number to actual characters sent
    shift_numbers = ['!', '@', '#', '$', '%', '^', '&', '*', '(', ')']
    
    for i in range(min(self.num_panes, len(shift_numbers))):
        key = shift_numbers[i]  # !, @, #, $, etc.
        cmd = f"resize-pane -Z -t {self.session_name}.{i}"
        self.bind_key_raw(key, cmd, table="prefix")
        
  #########################################
  def bind_key_raw(self, key=None, cmd=None, table="prefix"):
    """Bind a key to a raw command without modification"""
    assert(key is not None and cmd is not None)
    cmd_parts = cmd.split() if isinstance(cmd, str) else cmd
    self.post_chain.append([
      "tmux", "bind-key",
      "-T", table,
      key] + cmd_parts)

  #########################################
  def bind_key(self, key=None, cmd=None, table="prefix" ):
    """Bind a key to a command in the tmux session."""
    assert( key is not None and cmd is not None )
    # Handle commands with arguments (like select-layout even-vertical)
    if isinstance(cmd, list):
      cmd_parts = cmd
    else:
      cmd_parts = cmd.split()

    # For select-layout, we need to insert -t session_name before the layout name
    if len(cmd_parts) >= 2 and cmd_parts[0] == "select-layout":
      cmd_with_target = [cmd_parts[0], "-t", f"{self.session_name}"] + cmd_parts[1:]
    else:
      cmd_with_target = cmd_parts + ["-t", f"{self.session_name}"]

    self.post_chain.append([
      "tmux", "bind-key",
      "-T", table,
      key] + cmd_with_target)
      
  #########################################
  def kill(self):
    try:
      command.run(self.kill_session, check=False)
    except:
      pass
      
  #########################################
  def first_command(self, cmd):
    self.cmd_chain.add(self.cmd_new_session + cmd)
    
  #########################################
  def next_command(self, cmd):
    self.cmd_chain.add(self.cmd_add_session + cmd)
    self.num_panes += 1  # Increment when we actually add a pane
    
  #########################################
  def command(self, cmd):
    if self.cmd_chain.count == 0:
      self.first_command(cmd)
    else:
      self.next_command(cmd)
      
  #########################################
  def select_layout(self):
    self.cmd_chain.add(self.cmd_select_layout)
    
  #########################################
  def attach_session(self):
    self.cmd_chain.add(self.cmd_attach_session)
    
  #########################################
  def execute(self):
    self.bind_zoom_panes()
    ########################
    if self.kill_first:
      self.kill()
    ########################
    for item in self.post_chain:
      self.cmd_chain.add(item)
    ########################
    self.select_layout()
    self.attach_session()
    ########################
    OK = (self.cmd_chain.execute() == 0)
    return OK